@@ -85,7 +85,7 @@ Sl.setup <- function(G) {
       ## overlap testing requires the block ranges 
       for (j in 1:m) { ## get block range for each S[[j]]
         ir <- range((1:nb)[rowSums(abs(Sl[[b]]$S[[j]]))>0])
-        ic <- range((1:nb)[colSums(abs(Sl[[b]]$S[[j]]))>0])
+        #ic <- range((1:nb)[colSums(abs(Sl[[b]]$S[[j]]))>0]) ## symmetric not needed
         sbStart[j] <- ir[1];sbStop[j] <- ir[2] ## individual ranges
       } 
       split.ok <- TRUE
@@ -166,6 +166,7 @@ Sl.setup <- function(G) {
       ind <- 1:Sl[[b]]$rank
       for (j in 1:length(Sl[[b]]$S)) { ## project penalties into range space of total penalty
         Sl[[b]]$S[[j]] <- t(U[,ind])%*%Sl[[b]]$S[[j]]%*%U[,ind]
+        Sl[[b]]$S[[j]] <- (t(Sl[[b]]$S[[j]]) + Sl[[b]]$S[[j]])/2 ## avoid over-zealous chol sym check
         Sl[[b]]$rS[[j]] <- mroot(Sl[[b]]$S[[j]],Sl[[b]]$rank)
       }
       Sl[[b]]$ind <- rep(FALSE,ncol(U))
@@ -180,6 +181,7 @@ Sl.setup <- function(G) {
         St <- St + Sl[[b]]$S[[j]]/S.norm
         lambda <- c(lambda,1/S.norm)
       } 
+      St <- (t(St) + St)/2  ## avoid over-zealous chol sym check
       St <- t(mroot(St,Sl[[b]]$rank))
       indc <- Sl[[b]]$start:(Sl[[b]]$start+ncol(St)-1)
       indr <- Sl[[b]]$start:(Sl[[b]]$start+nrow(St)-1)
@@ -191,7 +193,7 @@ Sl.setup <- function(G) {
   Sl ## the penalty list
 } ## end of Sl.setup
 
-Sl.initial.repara <- function(Sl,X,inverse=FALSE,both.sides=TRUE,cov=TRUE) {
+Sl.initial.repara <- function(Sl,X,inverse=FALSE,both.sides=TRUE,cov=TRUE,nt=1) {
 ## Routine to apply initial Sl re-parameterization to model matrix X,
 ## or, if inverse==TRUE, to apply inverse re-para to parameter vector 
 ## or cov matrix. if inverse is TRUE and both.sides=FALSE then 
@@ -203,8 +205,10 @@ Sl.initial.repara <- function(Sl,X,inverse=FALSE,both.sides=TRUE,cov=TRUE) {
         for (b in 1:length(Sl)) { 
           ind <- Sl[[b]]$start:Sl[[b]]$stop
           if (is.matrix(Sl[[b]]$D)) { 
-            if (both.sides) X[ind,] <- Sl[[b]]$D%*%X[ind,,drop=FALSE]
-            X[,ind] <- X[,ind,drop=FALSE]%*%t(Sl[[b]]$D) 
+            if (both.sides) X[ind,] <- if (nt==1) Sl[[b]]$D%*%X[ind,,drop=FALSE] else 
+                                       pmmult(Sl[[b]]$D,X[ind,,drop=FALSE],FALSE,FALSE,nt=nt)
+            X[,ind] <- if (nt==1) X[,ind,drop=FALSE]%*%t(Sl[[b]]$D) else
+                       pmmult(X[,ind,drop=FALSE],Sl[[b]]$D,FALSE,TRUE,nt=nt)
           } else { ## Diagonal D
             X[,ind] <- t(Sl[[b]]$D * t(X[,ind,drop=FALSE]))
             if (both.sides) X[ind,] <- Sl[[b]]$D * X[ind,,drop=FALSE]
@@ -215,8 +219,10 @@ Sl.initial.repara <- function(Sl,X,inverse=FALSE,both.sides=TRUE,cov=TRUE) {
           ind <- Sl[[b]]$start:Sl[[b]]$stop
           if (is.matrix(Sl[[b]]$D)) { 
             Di <- if(is.null(Sl[[b]]$Di)) t(Sl[[b]]$D) else Sl[[b]]$Di
-            if (both.sides) X[ind,] <- t(Di)%*%X[ind,,drop=FALSE]
-            X[,ind] <- X[,ind,drop=FALSE]%*%Di 
+            if (both.sides) X[ind,] <- if (nt==1) t(Di)%*%X[ind,,drop=FALSE] else
+                            pmmult(Di,X[ind,,drop=FALSE],TRUE,FALSE,nt=nt)
+            X[,ind] <- if (nt==1) X[,ind,drop=FALSE]%*%Di else
+                       pmmult(X[,ind,drop=FALSE],Di,FALSE,FALSE,nt=nt)
           } else { ## Diagonal D
             Di <- 1/Sl[[b]]$D
             X[,ind] <- t(Di * t(X[,ind,drop=FALSE]))
@@ -233,15 +239,72 @@ Sl.initial.repara <- function(Sl,X,inverse=FALSE,both.sides=TRUE,cov=TRUE) {
     }
   } else for (b in 1:length(Sl)) { ## model matrix re-para
     ind <- Sl[[b]]$start:Sl[[b]]$stop
-    if (is.matrix(Sl[[b]]$D)) X[,ind] <- X[,ind,drop=FALSE]%*%Sl[[b]]$D else 
-    X[,ind] <- t(Sl[[b]]$D*t(X[,ind,drop=FALSE])) ## X[,ind]%*%diag(Sl[[b]]$D)
+    if (is.matrix(X)) { 
+      if (is.matrix(Sl[[b]]$D)) { 
+        if (both.sides)  X[ind,] <- if (nt==1) t(Sl[[b]]$D)%*%X[ind,,drop=FALSE] else 
+                         pmmult(Sl[[b]]$D,X[ind,,drop=FALSE],TRUE,FALSE,nt=nt)
+        X[,ind] <- if (nt==1) X[,ind,drop=FALSE]%*%Sl[[b]]$D else
+                   pmmult(X[,ind,drop=FALSE],Sl[[b]]$D,FALSE,FALSE,nt=nt)
+      } else { 
+        if (both.sides) X[ind,] <- Sl[[b]]$D * X[ind,,drop=FALSE]
+        X[,ind] <- t(Sl[[b]]$D*t(X[,ind,drop=FALSE])) ## X[,ind]%*%diag(Sl[[b]]$D)
+      }
+    } else {
+      if (is.matrix(Sl[[b]]$D)) X[ind] <- t(Sl[[b]]$D)%*%X[ind] else 
+      X[ind] <- Sl[[b]]$D*X[ind] 
+    }
   }
   X
 } ## end Sl.initial.repara
 
-ldetS <- function(Sl,rho,fixed,np,root=FALSE) {
+
+ldetSblock <- function(rS,rho,deriv=2,root=FALSE,nt=1) {
+## finds derivatives wrt rho of log|S| where 
+## S = sum_i tcrossprod(rS[[i]]*exp(rho[i]))
+## when S is full rank +ve def and no 
+## reparameterization is required....
+  lam <- exp(rho)
+  S <- pcrossprod(rS[[1]],trans=TRUE,nt=nt)*lam[1]
+      ##tcrossprod(rS[[1]])*lam[1] ## parallel
+  p <- ncol(S)
+  m <- length(rS)
+  if (m > 1) for (i in 2:m) S <- S + pcrossprod(rS[[i]],trans=TRUE,nt=nt)*lam[i]
+  ## S <- S + tcrossprod(rS[[i]])*lam[i] ## parallel
+  if (!root) E <- S
+  d <- diag(S);d[d<=0] <- 1;d <- sqrt(d)
+  S <- t(S/d)/d ## diagonally pre-condition
+  R <- if (nt>1) pchol(S,nt) else suppressWarnings(chol(S,pivot=TRUE))
+  piv <- attr(R,"pivot")
+  r <- attr(R,"rank")
+  if (r<p) R[(r+1):p,(r+1):p] <- 0 ## fix chol bug
+  if (root) {
+    rp <- piv;rp[rp] <- 1:p ## reverse pivot
+    E <- t(t(R[,rp])*d)
+  } 
+  if (r<p) { ## rank deficiency
+    R <- R[1:r,1:r]
+    piv <- piv[1:r]
+  }
+  RrS <- list();dS1 <- rep(0,m);dS2 <- matrix(0,m,m)
+  ## use dlog|S|/drhoi = lam_i tr(S^{-1}S_i) = tr(R^{-T}rS[[i]]rS[[i]]R^{-1} etc...
+  for (i in 1:m) {
+    RrS[[i]] <- pforwardsolve(R,rS[[i]][piv,]/d[piv],nt=nt) ## note R transposed internally - unlike forwardsolve!!
+    dS1[i] <- sum(RrS[[i]]^2)*lam[i] ## dlog|S|/drhoi
+    if (deriv==2) { 
+      RrS[[i]] <- pcrossprod(RrS[[i]],trans=TRUE,nt=nt)
+      #tcrossprod(RrS[[i]]) ## parallel
+      for (j in 1:i) {
+        dS2[i,j] <- dS2[j,i] <- -sum(RrS[[i]]*RrS[[j]])*lam[i]*lam[j]
+      }
+      dS2[i,i] <- dS2[i,i] + dS1[i]
+    }
+  }
+  list(det = 2*sum(log(diag(R))+log(d[piv])),det1=dS1,det2=dS2,E=E)
+} ## ldetSblock
+
+ldetS <- function(Sl,rho,fixed,np,root=FALSE,repara=TRUE,nt=1) {
 ## Get log generalized determinant of S stored blockwise in an Sl list.
-## Any multi-term blocks will be re-parameterized using gam.reparam, and
+## If repara=TRUE multi-term blocks will be re-parameterized using gam.reparam, and
 ## a re-parameterization object supplied in the returned object.
 ## rho contains log smoothing parameters, fixed is an array indicating whether they 
 ## are fixed (or free). np is the number of coefficients. root indicates
@@ -276,7 +339,8 @@ ldetS <- function(Sl,rho,fixed,np,root=FALSE) {
       ind <- k.sp:(k.sp+m-1) ## index for smoothing parameters
       ## call gam.reparam to deal with this block
       ## in a stable way...
-      grp <- gam.reparam(Sl[[b]]$rS,lsp=rho[ind],deriv=2) 
+      grp <- if (repara) gam.reparam(Sl[[b]]$rS,lsp=rho[ind],deriv=2) else 
+             ldetSblock(Sl[[b]]$rS,rho[ind],deriv=2,root=root,nt=nt)
       Sl[[b]]$lambda <- exp(rho[ind])
       ldS <- ldS + grp$det
       ## next deal with the derivatives...
@@ -289,25 +353,53 @@ ldetS <- function(Sl,rho,fixed,np,root=FALSE) {
         d2.ldS[dind,dind] <- grp$det2
         k.deriv <- k.deriv + nd
       }
-      ## now store the reparameterization information      
-      rp[[k.rp]] <- list(block =b,ind = (Sl[[b]]$start:Sl[[b]]$stop)[Sl[[b]]$ind],Qs = grp$Qs)
-      k.rp <- k.rp + 1
-      for (i in 1:m) {
-        Sl[[b]]$Srp[[i]] <- Sl[[b]]$lambda[i]*grp$rS[[i]]%*%t(grp$rS[[i]])
+      ## now store the reparameterization information   
+      if (repara) {   
+        rp[[k.rp]] <- list(block =b,ind = (Sl[[b]]$start:Sl[[b]]$stop)[Sl[[b]]$ind],Qs = grp$Qs)
+        k.rp <- k.rp + 1
+        for (i in 1:m) {
+          Sl[[b]]$Srp[[i]] <- Sl[[b]]$lambda[i]*grp$rS[[i]]%*%t(grp$rS[[i]])
+        }
       }
       k.sp <- k.sp + m
       if (root) { ## unpack the square root E'E
         ic <- Sl[[b]]$start:(Sl[[b]]$start+ncol(grp$E)-1)
         ir <- Sl[[b]]$start:(Sl[[b]]$start+nrow(grp$E)-1)
-        E[ir,ic] <- grp$E     
-        Sl[[b]]$St <- t(grp$E)%*%grp$E
-      }      
+        E[ir,ic] <- grp$E      
+        Sl[[b]]$St <- crossprod(grp$E)
+      } else {  
+        ## gam.reparam always returns root penalty in E, but 
+        ## ldetSblock returns penalty in E if root==FALSE
+        Sl[[b]]$St <- if (repara) crossprod(grp$E) else grp$E
+      }   
     } ## end of multi-S block branch
   } ## end of block loop
-  if (root) E <- E[rowSums(abs(E))!=0,] ## drop zero rows.
+  if (root) E <- E[rowSums(abs(E))!=0,,drop=FALSE] ## drop zero rows.
   list(ldetS=ldS,ldet1=d1.ldS,ldet2=d2.ldS,Sl=Sl,rp=rp,E=E)
 } ## end ldetS
 
+
+
+Sl.addS <- function(Sl,A,rho) {
+## Routine to add total penalty to matrix A. Sl is smooth penalty
+## list from Sl.setup, so initial reparameterizations have taken place,
+## and should have already been applied to A using Sl.initial.repara
+  k <- 1
+  for (b in 1:length(Sl)) {
+    ind <- (Sl[[b]]$start:Sl[[b]]$stop)[Sl[[b]]$ind] 
+    if (length(Sl[[b]]$S)==1) { ## singleton
+      diag(A)[ind] <-  diag(A)[ind] + exp(rho[k]) ## penalty is identity times sp
+      k <- k + 1
+    } else {
+      for (j in 1:length(Sl[[b]]$S)) {
+        A[ind,ind] <- A[ind,ind] + exp(rho[k]) * Sl[[b]]$S[[j]]
+        k <- k + 1
+      }
+    }
+  }
+  A
+} ## Sl.addS
+
 Sl.repara <- function(rp,X,inverse=FALSE,both.sides=TRUE) {
 ## Apply re-parameterization from ldetS to X, blockwise.
 ## If X is a matrix it is assumed to be a model matrix
@@ -379,12 +471,21 @@ Sl.mult <- function(Sl,A,k = 0,full=TRUE) {
             ind <- (Sl[[b]]$start:Sl[[b]]$stop)[Sl[[b]]$ind]
             if (full) { ## return zero answer with all zeroes in place
               B <- A*0
-              if (Amat) B[ind,] <- Sl[[b]]$Srp[[i]]%*%A[ind,] else  
-                        B[ind] <- Sl[[b]]$Srp[[i]]%*%A[ind]
+              if (Amat) {
+                B[ind,] <- if (is.null(Sl[[b]]$Srp)) Sl[[b]]$lambda[i]*(Sl[[b]]$S[[i]]%*%A[ind,]) else 
+                                                     Sl[[b]]$Srp[[i]]%*%A[ind,]
+              } else {
+                B[ind] <- if (is.null(Sl[[b]]$Srp)) Sl[[b]]$lambda[i]*(Sl[[b]]$S[[i]]%*%A[ind]) else 
+                                                    Sl[[b]]$Srp[[i]]%*%A[ind]
+              }
               A <- B
             } else { ## strip zero rows from answer
-              if (Amat) A <- Sl[[b]]$Srp[[i]]%*%A[ind,] else
-                        A <- as.numeric(Sl[[b]]$Srp[[i]]%*%A[ind])
+              if (is.null(Sl[[b]]$Srp)) {
+                A <- if (Amat)  Sl[[b]]$lambda[i]*(Sl[[b]]$S[[i]]%*%A[ind,]) else  
+                                Sl[[b]]$lambda[i]*as.numeric(Sl[[b]]$S[[i]]%*%A[ind])
+              } else {
+                A <- if (Amat) Sl[[b]]$Srp[[i]]%*%A[ind,] else as.numeric(Sl[[b]]$Srp[[i]]%*%A[ind])
+              }
             }
           }
           break
@@ -396,7 +497,7 @@ Sl.mult <- function(Sl,A,k = 0,full=TRUE) {
   A
 } ## end Sl.mult
 
-Sl.termMult <- function(Sl,A,full=FALSE) {
+Sl.termMult <- function(Sl,A,full=FALSE,nt=1) {
 ## returns a list containing the product of each element S of Sl
 ## with A. If full==TRUE then the results include the zero rows
 ## otherwise these are stripped out, but in that case each element 
@@ -426,12 +527,30 @@ Sl.termMult <- function(Sl,A,full=FALSE) {
         k <- k + 1
         if (full) { ## return answer with all zeroes in place
           B <- A*0
-          if (Amat) B[ind,] <- Sl[[b]]$Srp[[i]]%*%A[ind,,drop=FALSE] else  
-                    B[ind] <- Sl[[b]]$Srp[[i]]%*%A[ind]
+          if (is.null(Sl[[b]]$Srp)) {
+            if (Amat) { 
+              B[ind,] <- if (nt==1)  Sl[[b]]$lambda[i]*(Sl[[b]]$S[[i]]%*%A[ind,,drop=FALSE]) else 
+                         Sl[[b]]$lambda[i]*pmmult(Sl[[b]]$S[[i]],A[ind,,drop=FALSE],nt=nt) 
+            } else B[ind] <-  Sl[[b]]$lambda[i]*(Sl[[b]]$S[[i]]%*%A[ind])
+          } else {
+            if (Amat) { 
+              B[ind,] <- if (nt==1) Sl[[b]]$Srp[[i]]%*%A[ind,,drop=FALSE] else 
+                       pmmult(Sl[[b]]$Srp[[i]],A[ind,,drop=FALSE],nt=nt) 
+            } else B[ind] <- Sl[[b]]$Srp[[i]]%*%A[ind]
+          }
           SA[[k]] <- B
         } else { ## strip zero rows from answer
-          if (Amat) SA[[k]] <- Sl[[b]]$Srp[[i]]%*%A[ind,,drop=FALSE] else
-                    SA[[k]] <- as.numeric(Sl[[b]]$Srp[[i]]%*%A[ind])
+          if (is.null(Sl[[b]]$Srp)) {
+            if (Amat) {
+              SA[[k]] <- if (nt==1)  Sl[[b]]$lambda[i]*(Sl[[b]]$S[[i]]%*%A[ind,,drop=FALSE]) else
+                         Sl[[b]]$lambda[i]*pmmult(Sl[[b]]$S[[i]],A[ind,,drop=FALSE],nt=nt)
+            } else SA[[k]] <-  Sl[[b]]$lambda[i]*as.numeric(Sl[[b]]$S[[i]]%*%A[ind])
+          } else {
+            if (Amat) {
+              SA[[k]] <- if (nt==1) Sl[[b]]$Srp[[i]]%*%A[ind,,drop=FALSE] else
+                       pmmult(Sl[[b]]$Srp[[i]],A[ind,,drop=FALSE],nt=nt)
+            } else SA[[k]] <- as.numeric(Sl[[b]]$Srp[[i]]%*%A[ind])
+          }
           attr(SA[[k]],"ind") <- ind
         }
       } ## end of S loop for block b
@@ -440,10 +559,10 @@ Sl.termMult <- function(Sl,A,full=FALSE) {
   SA
 } ## end Sl.termMult
 
-d.detXXS <- function(Sl,PP) {
+d.detXXS <- function(Sl,PP,nt=1) {
 ## function to obtain derivatives of log |X'X+S| given unpivoted PP' where 
 ## P is inverse of R from the QR of the augmented model matrix. 
-  SPP <- Sl.termMult(Sl,PP,full=FALSE) ## SPP[[k]] is S_k PP'
+  SPP <- Sl.termMult(Sl,PP,full=FALSE,nt=nt) ## SPP[[k]] is S_k PP'
   nd <- length(SPP)
   d1 <- rep(0,nd);d2 <- matrix(0,nd,nd) 
   for (i in 1:nd) { 
@@ -462,39 +581,158 @@ Sl.ift <- function(Sl,R,X,y,beta,piv,rp) {
 ## function to obtain derviatives of \hat \beta by implicit differentiation
 ## and to use these directly to evaluate derivs of b'Sb and the RSS.
 ## piv and rp are the pivots and inverse pivots from the qr that produced R.
+## rssj and bSbj only contain the terms that will not cancel in rssj + bSbj
   beta <- beta[rp] ## unpivot
   Sb <- Sl.mult(Sl,beta,k = 0)          ## unpivoted
   Skb <- Sl.termMult(Sl,beta,full=TRUE) ## unpivoted
   rsd <- (X%*%beta - y)
-  Xrsd <- t(X)%*%rsd ## X'Xbeta - X'y
+  #Xrsd <- t(X)%*%rsd ## X'Xbeta - X'y
   nd <- length(Skb)
   np <- length(beta)
-  Sk.db <- db <- matrix(0,np,nd)
+  db <- matrix(0,np,nd)
   rss1 <- bSb1 <- rep(0,nd)  
   
   for (i in 1:nd) { ## compute the first derivatives
     db[,i] <- -backsolve(R,forwardsolve(t(R),Skb[[i]][piv]))[rp] ## d beta/ d rho_i
-    ##Sk.db[,i] <- Sl.mult(Sl,db[,i],k = i,full=TRUE)    ## S_i d beta/ d rho_i
-    rss1[i] <- 2 * sum(db[,i]*Xrsd)                    ## d rss / d rho_i
-    bSb1[i] <- 2 * sum(db[,i]*Sb) + sum(beta*Skb[[i]])       ## d b'Sb / d_rho_i
+    ## rss1[i] <- 0* 2 * sum(db[,i]*Xrsd)                      ## d rss / d rho_i
+    bSb1[i] <- sum(beta*Skb[[i]]) ## + 2 * sum(db[,i]*Sb)   ## d b'Sb / d_rho_i
   }
   XX.db <- t(X)%*%(X%*%db)
   S.db <- Sl.mult(Sl,db,k=0)
-  Sk.db <- Sl.termMult(Sl,db,full=TRUE) ## Sk.db[[j]][,k] is S_j d beta / d rho_k
+##  Sk.db <- Sl.termMult(Sl,db,full=TRUE) ## Sk.db[[j]][,k] is S_j d beta / d rho_k
 
   rss2 <- bSb2 <- matrix(0,nd,nd)
   for (k in 1:nd) { ## second derivative loop 
     for (j in k:nd) {
-      d2b <- (k==j)*db[,k] - backsolve(R,forwardsolve(t(R),Sk.db[[j]][piv,k]+Sk.db[[k]][piv,j]))[rp]
-      rss2[j,k] <- rss2[k,j] <- 2 * sum(d2b*Xrsd) + 2 * sum(db[,j]*XX.db[,k])
-      bSb2[j,k] <- bSb2[k,j] <- 2 * (sum(d2b*Sb) + sum(db[,k]*(Skb[[j]]+S.db[,j])) + sum(db[,j]*Skb[[k]])) +
-                                (k==j)*sum(beta*Skb[[k]])
+      ## d2b <- (k==j)*db[,k] - backsolve(R,forwardsolve(t(R),Sk.db[[j]][piv,k]+Sk.db[[k]][piv,j]))[rp]
+      rss2[j,k] <- rss2[k,j] <- 2 * sum(db[,j]*XX.db[,k]) ## + 2 * sum(d2b*Xrsd) 
+      bSb2[j,k] <- bSb2[k,j] <-  (k==j)*sum(beta*Skb[[k]])  + 2*(sum(db[,k]*(Skb[[j]]+S.db[,j])) + 
+                                 sum(db[,j]*Skb[[k]])) ## + 2 * (sum(d2b*Sb)   
+                               
     }
   }
-  list(rss =sum(rsd^2),bSb=sum(beta*Sb),rss1=rss1,bSb1=bSb1,rss2=rss2,bSb2=bSb2,d1b=db)
+  list(bSb=sum(beta*Sb),bSb1=bSb1,bSb2=bSb2,d1b=db,rss =sum(rsd^2),rss1=rss1,rss2=rss2)
 } ## end Sl.ift
 
-Sl.fit <- function(Sl,X,y,rho,fixed,log.phi=0,phi.fixed=TRUE,rss.extra=0,nobs=NULL,Mp=0) {
+Sl.iftChol <- function(Sl,XX,R,d,beta,piv) {
+## function to obtain derviatives of \hat \beta by implicit differentiation
+## and to use these directly to evaluate derivs of b'Sb and the RSS.
+## piv contains the pivots from the chol that produced R.
+## rssj and bSbj only contain the terms that will not cancel in rssj + bSbj
+  Sb <- Sl.mult(Sl,beta,k = 0)          ## unpivoted
+  Skb <- Sl.termMult(Sl,beta,full=TRUE) ## unpivoted
+  nd <- length(Skb)
+  np <- length(beta)
+  db <- matrix(0,np,nd)
+  rss1 <- bSb1 <- rep(0,nd)  
+  
+  for (i in 1:nd) { ## compute the first derivatives
+    db[piv,i] <- -backsolve(R,forwardsolve(t(R),Skb[[i]][piv]/d[piv]))/d[piv] ## d beta/ d rho_i
+    bSb1[i] <- sum(beta*Skb[[i]])  ## d b'Sb / d_rho_i
+  }
+  XX.db <- XX%*%db
+  #XX.db[piv,] <- d[piv]*(t(R)%*%(R%*%(d[piv]*db[piv,]))) ## X'Xdb
+
+  S.db <- Sl.mult(Sl,db,k=0)
+  ##Sk.db <- Sl.termMult(Sl,db,full=TRUE) ## Sk.db[[j]][,k] is S_j d beta / d rho_k
+
+  rss2 <- bSb2 <- matrix(0,nd,nd)
+  for (k in 1:nd) { ## second derivative loop 
+    for (j in k:nd) {
+      rss2[j,k] <- rss2[k,j] <- 2 * sum(db[,j]*XX.db[,k]) 
+      bSb2[j,k] <- bSb2[k,j] <-  (k==j)*sum(beta*Skb[[k]])  + 2*(sum(db[,k]*(Skb[[j]]+S.db[,j])) + 
+                                 sum(db[,j]*Skb[[k]]))                                  
+    }
+  }
+  list(bSb=sum(beta*Sb),bSb1=bSb1,bSb2=bSb2,d1b=db,rss1=rss1,rss2=rss2)
+} ## end Sl.iftChol
+
+
+Sl.fitChol <- function(Sl,XX,f,rho,yy=0,L=NULL,rho0=0,log.phi=0,phi.fixed=TRUE,nobs=0,Mp=0,nt=1,tol=0) {
+## given X'WX in XX and f=X'Wy solves the penalized least squares problem
+## with penalty defined by Sl and rho, and evaluates a REML Newton step, the REML 
+## gradiant and the the estimated coefs bhat. If phi.fixed=FALSE then we need 
+## yy = y'Wy in order to get derivsatives w.r.t. phi. 
+  
+  rho <- if (is.null(L)) rho + rho0 else L%*%rho + rho0
+
+  ## get log|S|_+ without stability transform... 
+  fixed <- rep(FALSE,length(rho))
+  ldS <- ldetS(Sl,rho,fixed,np=ncol(XX),root=FALSE,repara=FALSE,nt=nt)
+  
+  ## now the Choleki factor of the penalized Hessian... 
+  #XXp <- XX+crossprod(ldS$E) ## penalized Hessian
+  XXp <- Sl.addS(Sl,XX,rho)
+
+  d <- diag(XXp);ind <- d<=0
+  d[ind] <- 1;d[!ind] <- sqrt(d[!ind])
+  #XXp <- t(XXp/d)/d ## diagonally precondition
+  R <- if (nt>1) pchol(t(XXp/d)/d,nt) else suppressWarnings(chol(t(XXp/d)/d,pivot=TRUE))
+  r <- Rrank(R);p <- ncol(XXp)
+  piv <- attr(R,"pivot") #;rp[rp] <- 1:p
+  if (r<p) { ## drop rank deficient terms...
+    R <- R[1:r,1:r]
+    piv <- piv[1:r]
+  }
+  beta <- rep(0,p)
+  beta[piv] <- backsolve(R,(forwardsolve(t(R),f[piv]/d[piv])))/d[piv]
+ 
+  ## get component derivatives based on IFT...
+  dift <- Sl.iftChol(ldS$Sl,XX,R,d,beta,piv)
+ 
+  ## now the derivatives of log|X'X+S|
+  P <- pbsi(R,nt=nt,copy=TRUE) ## invert R 
+  PP <- matrix(0,p,p)
+  PP[piv,piv] <- if (nt==1) tcrossprod(P) else pRRt(P,nt) ## PP'
+  PP <- t(PP/d)/d
+  ldetXXS <- 2*sum(log(diag(R))+log(d[piv])) ## log|X'X+S|
+  dXXS <- d.detXXS(ldS$Sl,PP,nt=nt) ## derivs of log|X'X+S|
+
+  phi <- exp(log.phi)  
+
+  reml1 <-  (dXXS$d1[!fixed] - ldS$ldet1 + 
+            (dift$rss1[!fixed] + dift$bSb1[!fixed])/phi)/2
+
+  reml2 <- (dXXS$d2[!fixed,!fixed] - ldS$ldet2 +  
+           (dift$rss2[!fixed,!fixed] + dift$bSb2[!fixed,!fixed])/phi)/2 
+ 
+  if (!phi.fixed) {
+    n <- length(reml1)
+    rss.bSb <- yy - sum(beta*f) ## use identity ||y-Xb|| + b'Sb = y'y - b'X'y (b is minimizer)
+    reml1[n+1] <- (-rss.bSb/phi + nobs - Mp)/2
+    d <- c(-(dift$rss1[!fixed] + dift$bSb1[!fixed]),rss.bSb)/(2*phi)
+    reml2 <- rbind(cbind(reml2,d[1:n]),d)
+  }
+
+  if (!is.null(L)) {
+    reml1 <- t(L)%*%reml1
+    reml2 <- t(L)%*%reml2%*%L
+  }
+  uconv.ind <- (abs(reml1) > tol)|(abs(diag(reml2))>tol)
+  hess <- reml2
+  grad <- reml1
+  if (sum(uconv.ind)!=ncol(reml2)) { 
+    reml1 <- reml1[uconv.ind]
+    reml2 <- reml2[uconv.ind,uconv.ind]
+  }
+
+  er <- eigen(reml2,symmetric=TRUE)
+  er$values <- abs(er$values)
+  me <- max(er$values)*.Machine$double.eps^.5
+  er$values[er$values<me] <- me
+  step <- rep(0,length(uconv.ind))
+  step[uconv.ind] <- -er$vectors%*%((t(er$vectors)%*%reml1)/er$values)
+
+  ## limit the step length...
+  ms <- max(abs(step))
+  if (ms>4) step <- 4*step/ms
+
+  ## return the coefficient estimate, the reml grad and the Newton step...
+  list(beta=beta,grad=grad,step=step,db=dift$d1b,PP=PP,R=R,piv=piv,rank=r,
+       hess=hess,ldetS=ldS$ldetS,ldetXXS=ldetXXS)
+} ## Sl.fitChol
+
+Sl.fit <- function(Sl,X,y,rho,fixed,log.phi=0,phi.fixed=TRUE,rss.extra=0,nobs=NULL,Mp=0,nt=1) {
 ## fits penalized regression model with model matrix X and 
 ## initialised block diagonal penalty Sl to data in y, given 
 ## log smoothing parameters rho. 
@@ -504,11 +742,11 @@ Sl.fit <- function(Sl,X,y,rho,fixed,log.phi=0,phi.fixed=TRUE,rss.extra=0,nobs=NU
   phi <- exp(log.phi)
   if (is.null(nobs)) nobs <- n
   ## get log|S|_+ stably...
-  ldS <- ldetS(Sl,rho,fixed,np,root=TRUE)
+  ldS <- ldetS(Sl,rho,fixed,np,root=TRUE,nt=nt)
   ## apply resulting stable re-parameterization to X...
   X <- Sl.repara(ldS$rp,X)
-  ## get pivoted QR decomp of augmented model matrix
-  qrx <- qr(rbind(X,ldS$E),LAPACK=TRUE)
+  ## get pivoted QR decomp of augmented model matrix (in parallel if nt>1)
+  qrx <- if (nt>1) pqr2(rbind(X,ldS$E),nt=nt) else qr(rbind(X,ldS$E),LAPACK=TRUE)
   rp <- qrx$pivot;rp[rp] <- 1:np ## reverse pivot vector
   ## find pivoted \hat beta...
   R <- qr.R(qrx)
@@ -518,22 +756,26 @@ Sl.fit <- function(Sl,X,y,rho,fixed,log.phi=0,phi.fixed=TRUE,rss.extra=0,nobs=NU
   ## get component derivatives based on IFT...
   dift <- Sl.ift(ldS$Sl,R,X,y,beta,qrx$pivot,rp)
   ## and the derivatives of log|X'X+S|...
-  P <- backsolve(R,diag(np))[rp,] ## invert R and row unpivot
-  PP <- tcrossprod(P) ## PP'
+  P <- pbsi(R,nt=nt,copy=TRUE) ## invert R 
+  ## P <- backsolve(R,diag(np))[rp,] ## invert R and row unpivot
+  ## crossprod and unpivot (don't unpivot if unpivoting P above)
+  PP <- if (nt==1) tcrossprod(P)[rp,rp] else pRRt(P,nt)[rp,rp] ## PP'
   ldetXXS <- 2*sum(log(abs(diag(R)))) ## log|X'X+S|
-  dXXS <- d.detXXS(ldS$Sl,PP) ## derivs of log|X'X+S|
+  dXXS <- d.detXXS(ldS$Sl,PP,nt=nt) ## derivs of log|X'X+S|
   ## all ingredients are now in place to form REML score and 
   ## its derivatives....
   reml <- (rss.bSb/phi + (nobs-Mp)*log(2*pi*phi) +
            ldetXXS - ldS$ldetS)/2
-  reml1 <- ((dift$rss1[!fixed] + dift$bSb1[!fixed])/phi + 
-            dXXS$d1[!fixed] - ldS$ldet1)/2
-  reml2 <- ((dift$rss2[!fixed,!fixed] + dift$bSb2[!fixed,!fixed])/phi + 
-           dXXS$d2[!fixed,!fixed] - ldS$ldet2)/2
+  reml1 <-  (dXXS$d1[!fixed] - ldS$ldet1 + # dift$bSb1[!fixed]/phi)/2 
+            (dift$rss1[!fixed] + dift$bSb1[!fixed])/phi)/2
+
+  reml2 <- (dXXS$d2[!fixed,!fixed] - ldS$ldet2 + #dift$bSb2[!fixed,!fixed]/phi)/2 
+           (dift$rss2[!fixed,!fixed] + dift$bSb2[!fixed,!fixed])/phi)/2 
   ## finally add in derivatives w.r.t. log.phi
   if (!phi.fixed) {
     n <- length(reml1)
     reml1[n+1] <- (-rss.bSb/phi + nobs - Mp)/2
+    #d <- c(-(dift$bSb1[!fixed]),rss.bSb)/(2*phi)
     d <- c(-(dift$rss1[!fixed] + dift$bSb1[!fixed]),rss.bSb)/(2*phi)
     reml2 <- rbind(cbind(reml2,d[1:n]),d)
   } 
@@ -547,7 +789,7 @@ Sl.fit <- function(Sl,X,y,rho,fixed,log.phi=0,phi.fixed=TRUE,rss.extra=0,nobs=NU
 } ## Sl.fit
 
 fast.REML.fit <- function(Sl,X,y,rho,L=NULL,rho.0=NULL,log.phi=0,phi.fixed=TRUE,
-                 rss.extra=0,nobs=NULL,Mp=0,conv.tol=.Machine$double.eps^.5) {
+                 rss.extra=0,nobs=NULL,Mp=0,conv.tol=.Machine$double.eps^.5,nt=1) {
 ## estimates log smoothing parameters rho, by optimizing fast REML 
 ## using Newton's method. On input Sl is a block diagonal penalty 
 ## structure produced by Sl.setup, while X is a model matrix 
@@ -556,11 +798,11 @@ fast.REML.fit <- function(Sl,X,y,rho,L=NULL,rho.0=NULL,log.phi=0,phi.fixed=TRUE,
 ## structurally un-identifiable coefficients. 
 ## Note that lower bounds on smoothing parameters are not handled.
   maxNstep <- 5  
-
+  
   if (is.null(nobs)) nobs <- nrow(X)
   np <- ncol(X)
   if (nrow(X) > np) { ## might as well do an initial QR step
-    qrx <- qr(X,LAPACK=TRUE)
+    qrx <- if (nt>1) pqr2(X,nt=nt) else qr(X,LAPACK=TRUE)
     rp <- qrx$pivot
     rp[rp] <- 1:np
     X <- qr.R(qrx)[,rp]
@@ -580,7 +822,7 @@ fast.REML.fit <- function(Sl,X,y,rho,L=NULL,rho.0=NULL,log.phi=0,phi.fixed=TRUE,
   fixed <- rep(FALSE,nrow(L))
  
   
-  best <- Sl.fit(Sl,X,y,L%*%rho+rho.0,fixed,log.phi,phi.fixed,rss.extra,nobs,Mp)
+  best <- Sl.fit(Sl,X,y,L%*%rho+rho.0,fixed,log.phi,phi.fixed,rss.extra,nobs,Mp,nt=nt)
   ## get a typical scale for the reml score... 
   reml.scale <- abs(best$reml) + best$rss/best$nobs
  
@@ -624,14 +866,14 @@ fast.REML.fit <- function(Sl,X,y,rho,L=NULL,rho.0=NULL,log.phi=0,phi.fixed=TRUE,
     step[uconv.ind] <- uc.step ## step includes converged
     ## try out the step...
     rho1 <- L%*%(rho + step)+rho.0; if (!phi.fixed) log.phi <- rho1[nr+1]
-    trial <- Sl.fit(Sl,X,y,rho1[1:nr],fixed,log.phi,phi.fixed,rss.extra,nobs,Mp)
+    trial <- Sl.fit(Sl,X,y,rho1[1:nr],fixed,log.phi,phi.fixed,rss.extra,nobs,Mp,nt=nt)
     k <- 0
     while (trial$reml>best$reml && k<35) { ## step half until improvement
       step <- step/2;k <- k + 1
       rho1 <- L%*%(rho + step)+rho.0; if (!phi.fixed) log.phi <- rho1[nr+1]
-      trial <- Sl.fit(Sl,X,y,rho1[1:nr],fixed,log.phi,phi.fixed,rss.extra,nobs,Mp)
+      trial <- Sl.fit(Sl,X,y,rho1[1:nr],fixed,log.phi,phi.fixed,rss.extra,nobs,Mp,nt=nt)
     }
-    if (k==35 && trial$reml>best$reml) { ## step has failed
+    if ((k==35 && trial$reml>best$reml)||(sum(rho != rho + step)==0)) { ## step has failed
       step.failed <- TRUE
       break ## can get no further
     }
@@ -667,7 +909,7 @@ fast.REML.fit <- function(Sl,X,y,rho,L=NULL,rho.0=NULL,log.phi=0,phi.fixed=TRUE,
   best ## return the best fit (note that it will need post-processing to be useable)
 } ## end fast.REML.fit
 
-ident.test <- function(X,E) {
+ident.test <- function(X,E,nt=1) {
 ## routine to identify structurally un-identifiable coefficients
 ## for model with model matrix X and scaled sqrt penalty matrix E
 ## lambda is smoothing parameter vector corresponding to E, 
@@ -679,7 +921,7 @@ ident.test <- function(X,E) {
 ## then beta.full[undrop] <- beta, is the full, zero padded 
 ## coeff vector, with dropped coefs re-nstated as zeroes. 
   Xnorm <- norm(X,type="F")
-  qrx <- qr(rbind(X/Xnorm,E),LAPACK=TRUE) ## pivoted QR
+  qrx <- if (nt>1) pqr2(rbind(X/Xnorm,E),nt=nt) else qr(rbind(X/Xnorm,E),LAPACK=TRUE) ## pivoted QR
   rank <- Rrank(qr.R(qrx),tol=.Machine$double.eps^.75)
   drop <- qrx$pivot[-(1:rank)] ## index of un-identifiable coefs
   undrop <- 1:ncol(X) 
@@ -745,13 +987,13 @@ Sl.drop <- function(Sl,drop,np) {
   Sl
 } ## Sl.drop
 
-Sl.Xprep <- function(Sl,X) { 
+Sl.Xprep <- function(Sl,X,nt=1) { 
 ## Sl is block diag object from Sl.setup, X is a model matrix
 ## this routine applies preliminary Sl transformations to X
 ## tests for structural identifibility problems and drops
 ## un-identifiable parameters.
-  X <- Sl.initial.repara(Sl,X) ## apply re-para used in Sl to X
-  id <- ident.test(X,attr(Sl,"E")) ## deal with structural identifiability
+  X <- Sl.initial.repara(Sl,X,inverse=FALSE,both.sides=FALSE,cov=FALSE,nt=nt) ## apply re-para used in Sl to X
+  id <- ident.test(X,attr(Sl,"E"),nt=nt) ## deal with structural identifiability
   ## id contains drop, undrop, lambda
   if (length(id$drop)>0) { ## then there is something to do here 
     Sl <- Sl.drop(Sl,id$drop,ncol(X)) ## drop unidentifiable from Sl
@@ -764,7 +1006,7 @@ Sl.Xprep <- function(Sl,X) {
 } ## end Sl.Xprep
 
 
-Sl.postproc <- function(Sl,fit,undrop,X0,cov=FALSE,scale = -1) {
+Sl.postproc <- function(Sl,fit,undrop,X0,cov=FALSE,scale = -1,L,nt=nt) {
 ## reverse the various fitting re-parameterizations.
 ## X0 is the orginal model matrix before any re-parameterization
 ## or parameter dropping. Sl is also the original *before parameter 
@@ -772,16 +1014,17 @@ Sl.postproc <- function(Sl,fit,undrop,X0,cov=FALSE,scale = -1) {
   np <- ncol(X0)
   beta <- rep(0,np)
   beta[undrop] <- Sl.repara(fit$rp,fit$beta,inverse=TRUE)
-  beta <- Sl.initial.repara(Sl,beta,inverse=TRUE)
+  beta <- Sl.initial.repara(Sl,beta,inverse=TRUE,both.sides=TRUE,cov=TRUE,nt=nt)
  
   if (cov) { 
     d1b <- matrix(0,np,ncol(fit$d1b))
     ## following construction a bit ugly due to Sl.repara assumptions...
     d1b[undrop,] <- t(Sl.repara(fit$rp,t(fit$d1b),inverse=TRUE,both.sides=FALSE))
-    for (i in 1:ncol(d1b)) d1b[,i] <- Sl.initial.repara(Sl,as.numeric(d1b[,i]),inverse=TRUE) ## d beta / d rho matrix
+    for (i in 1:ncol(d1b)) d1b[,i] <- 
+        Sl.initial.repara(Sl,as.numeric(d1b[,i]),inverse=TRUE,both.sides=TRUE,cov=TRUE,nt=nt) ## d beta / d rho matrix
     PP <- matrix(0,np,np)
     PP[undrop,undrop] <-  Sl.repara(fit$rp,fit$PP,inverse=TRUE)
-    PP <- Sl.initial.repara(Sl,PP,inverse=TRUE)
+    PP <- Sl.initial.repara(Sl,PP,inverse=TRUE,both.sides=TRUE,cov=TRUE,nt=nt)
     #XPP <- crossprod(t(X0),PP)*X0
     #hat <- rowSums(XPP);edf <- colSums(XPP)
     XPP <- crossprod(t(X0),PP)
@@ -790,9 +1033,13 @@ Sl.postproc <- function(Sl,fit,undrop,X0,cov=FALSE,scale = -1) {
     edf <- diag(F)
     edf1 <- 2*edf - rowSums(t(F)*F) 
     ## edf <- rowSums(PP*crossprod(X0)) ## diag(PP%*%(t(X0)%*%X0))
-    if (scale<=0) scale <- fit$rss/(fit$nobs - sum(edf))
+    if (scale<=0) { 
+      scale <- fit$rss/(fit$nobs - sum(edf))
+    }
     Vp <- PP * scale ## cov matrix
     ## sp uncertainty correction... 
+    ## BUG: possibility of L ignored here.
+    if (!is.null(L)) d1b <- d1b%*%L
     M <- ncol(d1b) 
     ev <- eigen(fit$outer.info$hess,symmetric=TRUE)
     ind <- ev$values <= 0
