@@ -1,7 +1,7 @@
 #  File src/library/stats/R/acf.R
 #  Part of the R package, http://www.R-project.org
 #
-#  Copyright (C) 1999-2012 The R Core Team
+#  Copyright (C) 1999-2015 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -24,7 +24,8 @@ acf <-
     type <- match.arg(type)
     if(type == "partial") {
         m <- match.call()
-        m[[1L]] <- as.name("pacf")
+        ## need stats:: for non-standard evaluation
+        m[[1L]] <- quote(stats::pacf)
         m$type <- NULL
         return(eval(m, parent.frame()))
     }
@@ -75,7 +76,7 @@ pacf.default <- function(x, lag.max = NULL, plot = TRUE,
     if (lag.max < 1) stop("'lag.max' must be at least 1")
 
     if(is.matrix(x)) {
-        if(any(is.na(x))) stop("NAs in 'x'")
+        if(anyNA(x)) stop("NAs in 'x'")
         nser <- ncol(x)
         x <- sweep(x, 2, colMeans(x), check.margin=FALSE)
         lag <- matrix(1, nser, nser)
@@ -173,9 +174,7 @@ plot.acf <-
         iind <- (I-1)*nr + 1L:nr
         jind <- (J-1)*nr + 1L:nr
         if(verbose)
-            message("Page [",I,",",J,"]: i =",
-                    paste(iind,collapse=","),"; j =",
-                    paste(jind,collapse=","), domain = NA)
+            message(gettextf("Page [%d,%d]: i =%s; j =%s", I, J, paste(iind,collapse=","), paste(jind,collapse=",")), domain = NA)
         for (i in iind) for (j in jind)
             if(max(i,j) > nser) {
                 frame(); box(col = "light gray")
