@@ -2,7 +2,7 @@
 #  Part of the R package, http://www.R-project.org
 #
 #  Copyright (C) 1999-1999 Saikat DebRoy, Douglas M. Bates, Jose C. Pinheiro
-#  Copyright (C) 2000-2013 The R Core Team
+#  Copyright (C) 2000-2015 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -119,11 +119,8 @@ nlsModel.plinear <- function(form, data, start, wts)
         dtdot <- function(A, b) apply(A, MARGIN = c(2L,3L), FUN = "%*%", b)
     }
 
-    getPars.noVarying <- function()
-        unlist(setNames(lapply(names(ind), get, envir = env), names(ind)))
-    getPars.varying <- function()
-        unlist(setNames(lapply(names(ind), get, envir = env),
-                        names(ind)))[useParams]
+    getPars.noVarying <- function() unlist(mget(names(ind), env))
+    getPars.varying   <- function() unlist(mget(names(ind), env))[useParams]
     getPars <- getPars.noVarying
 
     internalPars <- getPars()
@@ -235,8 +232,7 @@ nlsModel <- function(form, data, start, wts, upper=NULL)
         ind[[i]] <- parLength + seq_along(start[[i]])
         parLength <- parLength + length(start[[i]])
     }
-    getPars.noVarying <- function()
-        unlist(setNames(lapply(names(ind), get, envir = env), names(ind)))
+    getPars.noVarying <- function() unlist(mget(names(ind), env))
     getPars <- getPars.noVarying
     internalPars <- getPars()
 
@@ -296,9 +292,7 @@ nlsModel <- function(form, data, start, wts, upper=NULL)
     if(QR$rank < qrDim)
         stop("singular gradient matrix at initial parameter estimates")
 
-    getPars.varying <- function()
-        unlist(setNames(lapply(names(ind), get, envir = env),
-                        names(ind)))[useParams]
+    getPars.varying <- function() unlist(mget(names(ind), env))[useParams]
     setPars.noVarying <- function(newPars)
     {
         assign("internalPars", newPars, envir = thisEnv)
@@ -419,7 +413,6 @@ nls_port_fit <- function(m, start, lower, upper, control, trace, give.v=FALSE)
     if (trace)
         iv[port_cpos[["trace"]]] <- 1L
     scale <- 1
-    low <- upp <- NULL
     if (any(lower != -Inf) || any(upper != Inf)) {
         low <- rep_len(as.double(lower), length(par))
         upp <- rep_len(as.double(upper), length(par))
@@ -427,7 +420,9 @@ nls_port_fit <- function(m, start, lower, upper, control, trace, give.v=FALSE)
             iv[1L] <- 300
 	    return(if(give.v) list(iv = iv, v = v[seq_len(18L)]) else iv)
         }
-    }
+    } else
+    	low <- upp <- numeric()
+    	
     if(p > 0) {
         ## driver routine port_nlsb() in ../src/port.c -- modifies m & iv
         .Call(C_port_nlsb, m,
@@ -564,6 +559,7 @@ nls <-
                            env = environment(formula))
             mf$start <- mf$control <- mf$algorithm <- mf$trace <- mf$model <- NULL
             mf$lower <- mf$upper <- NULL
+            ## need stats:: for non-standard evaluation
             mf[[1L]] <- quote(stats::model.frame)
             mf <- eval.parent(mf)
             n <- nrow(mf)
@@ -600,8 +596,11 @@ nls <-
     }
     ## Iterate
     if (algorithm != "port") {
-	if (!identical(lower, -Inf) || !identical(upper, +Inf))
+	if (!identical(lower, -Inf) || !identical(upper, +Inf)) {
 	    warning('upper and lower bounds ignored unless algorithm = "port"')
+	    cl$lower <- NULL # see PR#15960 -- confint() would use these regardless of algorithm
+	    cl$upper <- NULL
+	}
         convInfo <- .Call(C_nls_iter, m, ctrl, trace)
 	nls.out <- list(m = m, convInfo = convInfo,
 			data = substitute(data), call = cl)
