@@ -1,7 +1,7 @@
 #  File src/library/tools/R/Rd2latex.R
 #  Part of the R package, http://www.R-project.org
 #
-#  Copyright (C) 1995-2012 The R Core Team
+#  Copyright (C) 1995-2015 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -79,7 +79,7 @@ Rd2latex <- function(Rd, out="", defines=.Platform$OS.type, stages="render",
         } else {
             function(x, con, outputEncoding, ...) {
                 x <- iconv(x, "UTF-8", outputEncoding,  mark=FALSE)
-                if (any(is.na(x))) {
+                if (anyNA(x)) {
                     x <- iconv(x, "UTF-8", outputEncoding,
                                sub="byte", mark=FALSE)
                     encode_warn <<- TRUE
@@ -102,7 +102,7 @@ Rd2latex <- function(Rd, out="", defines=.Platform$OS.type, stages="render",
     }
 
     envTitles <- c("\\description"="Description", "\\usage"="Usage",
-        "\\synopsis"="Usage", "\\arguments"="Arguments",
+        "\\arguments"="Arguments",
         "\\format"="Format", "\\details"="Details", "\\note"="Note",
         "\\section"="", "\\author"="Author",
         "\\references"="References", "\\source"="Source",
@@ -111,7 +111,6 @@ Rd2latex <- function(Rd, out="", defines=.Platform$OS.type, stages="render",
 
     sectionExtras <-
     c("\\usage"="verbatim",
-      "\\synopsis"="verbatim",
       "\\arguments"="ldescription",
       "\\examples"="ExampleCode")
 
@@ -212,9 +211,13 @@ Rd2latex <- function(Rd, out="", defines=.Platform$OS.type, stages="render",
             url <- as.character(block[[1L]])
             tag <- "\\Rhref"
         }
-    	of0(tag, "{",
-            gsub("\n", "", paste(as.character(url), collapse="")),
-            "}")
+        ## cleanup URL
+        url <- trimws(gsub("\n", "",
+                           paste(as.character(url), collapse = ""),
+                           fixed = TRUE, useBytes = TRUE))
+        ## escape % for LaTeX
+        url <- gsub("%", "\\%",  url, fixed = TRUE, useBytes = TRUE)
+    	of0(tag, "{", url, "}")
         if (tag == "\\Rhref") {
             of1("{")
             writeContent(block[[2L]], tag)
@@ -230,8 +233,6 @@ Rd2latex <- function(Rd, out="", defines=.Platform$OS.type, stages="render",
             latex_link_trans0(parts$dest), "}")
     }
 
-    writeComment <- function(txt) of0(txt, '\n')
-
     writeDR <- function(block, tag) {
         if (length(block) > 1L) {
             of1('## Not run: ')
@@ -298,35 +299,6 @@ Rd2latex <- function(Rd, out="", defines=.Platform$OS.type, stages="render",
         x
     }
 
-    latex_code_trans  <- function(x)
-    {
-        BSL = '@BSL@';
-        LATEX_SPECIAL = '$^&~_#'
-        if(grepl(LATEX_SPECIAL, x)) {
-            x <- fsub("\\\\", BSL, x)
-            ## unescape (should not be escaped)
-            x <- psub("\\\\([$^&~_#])", "\\1", x)
-            x <- psub("[$^&~_#]", "\\1&", x) #- escape them
-            x <- fsub("^", "\\textasciicircum{}", x) # ^ is SPECIAL
-            x <- fsub("~", "\\textasciitilde{}", x)
-            x <- fsub(BSL, "\\bsl{}", x)
-            x <- fsub("\\", "\\bsl{}", x)
-        }
-        ## avoid conversion to guillemets
-        x <- fsub("<<", "<{}<", x)
-        x <- fsub(">>", ">{}>", x)
-        x <- fsub(",,", ",{},", x) # ,, is a ligature in the ae font.
-        x <- psub("\\\\bsl{}var\\\\{([^}]+)\\\\}", "\\var{\\1}", x)
-        x
-}
-
-    latex_link_trans <- function(x)
-    {
-        x <- fsub("<-.", "<\\Rdash.", x)
-        x <- psub("<-$", "<\\Rdash", x)
-        x
-    }
-
     latex_code_alias <- function(x)
     {
         x <- fsub("{", "\\{", x)
@@ -339,13 +311,6 @@ Rd2latex <- function(Rd, out="", defines=.Platform$OS.type, stages="render",
         x
     }
 
-    latex_code_aliasAA <- function(x)
-    {
-        x <- latex_code_trans(x)
-        x <- latex_link_trans(x)
-        psub("\\\\([!|])", '"\\1', x)
-    }
-
     currentAlias <- NA_character_
 
     writeAlias <- function(block, tag) {
@@ -401,7 +366,7 @@ Rd2latex <- function(Rd, out="", defines=.Platform$OS.type, stages="render",
                "\\option" =,
                "\\samp" = writeWrapped(block, tag),
                ## really verbatim
-                "\\url"=,
+               "\\url"=,
                "\\href"= writeURL(block, tag),
                ## R-like
                "\\code"= {
@@ -636,8 +601,6 @@ Rd2latex <- function(Rd, out="", defines=.Platform$OS.type, stages="render",
     Rdfile <- attr(Rd, "Rdfile")
     sections <- RdTags(Rd)
 
-    enc <- which(sections == "\\encoding")
-
     if (is.character(out)) {
         if(out == "") {
             con <- stdout()
