@@ -1,8 +1,8 @@
 #  File src/library/utils/R/completion.R
 #  Part of the R package, http://www.R-project.org
 #
-# Copyright (C) 2006  Deepayan Sarkar
-#               2006-2012  The R Core Team
+# Copyright     2006 Deepayan Sarkar
+#           (C) 2006-2015  The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -45,6 +45,91 @@
 ### careful because 0 length components in sprintf will cause errors.
 
 
+## [July 2013] First attempt to support fuzzy matching, if
+## rc.settings(fuzzy=TRUE), based on suggestion from Rasmus Baath.
+## Initially, this replaces use of grep() to find matches by
+## findMatches(), which behaves differently depending on the 'fuzzy'
+## setting.  This does not affect basic object name completion, which
+## is done using apropos().  For that, we need to write a fuzzy
+## version of apropos (which is not that difficult; just loop through
+## everything in the search path).
+
+findExactMatches <- function(pattern, values)
+{
+    grep(pattern, values, value = TRUE)
+}
+
+### agrep() version
+##
+## findFuzzyMatches <- function(pattern, values)
+## {
+##     ## Try exact matches first, and return them if found
+##     ans <- findExactMatches(pattern, values)
+##     if (length(ans) == 0) {
+##         fuzzies <-
+##             agrep(pattern, values, max.distance = 2,
+##                   ignore.case = TRUE, fixed = FALSE, value = TRUE)
+##         ## Multiple inconsistent matches will lead to more deletion
+##         ## than reasonable.  To avoid this, we find distances, and
+##         ## return the one with minimum distance.  However, if minimum
+##         ## distance is not unique, this will still delete.
+##         ## E.g., a = list(.foobar = 1, foo.bar = 2) ; a$foob<TAB>
+##         if (length(fuzzies) == 0) character(0)
+##         else {
+##             fdist <- adist(pattern, fuzzies, ignore.case=TRUE, partial = TRUE, fixed = FALSE)
+##             fmin <- which(fdist == min(fdist))
+##             fuzzies[fmin]
+##         }
+##     }
+##     else
+##         ans
+## }
+
+### normalizing version (from Rasmus Baath)
+##
+
+findFuzzyMatches <- function(pattern, values) {
+    ## FIXME: option to allow experimentation, remove eventually
+    if (!is.null(ffun <- getOption("fuzzy.match.fun"))) {
+        return (ffun(pattern, values))
+    }
+    ## Try exact matches first, and return them if found
+    exact.matches <- findExactMatches(pattern, values)
+    if (length(exact.matches) == 0) {
+        ## Removes "\\." and "_" in the pattern excluding the anchor
+        ## (^) and the first character but does not removes "\\." and
+        ## "_" if it is the last character.
+        normalizedPattern <- gsub("(?<!\\^)(?<=.)(\\\\\\.|_)(?!$)", "", pattern, perl = TRUE)
+        ## Replaces "\\." and "_" last in the pattern with ".", that
+        ## is the pattern is "seq\\."  we transform it into "seq." in
+        ## order to match "seq_along" and "seq.int" but not "seq".
+        normalizedPattern <- gsub("(\\\\\\.|_)$", ".", normalizedPattern)
+        normalizedValues <- gsub("(?<=.)[._]", "", values, perl = TRUE)
+        values[ grep(normalizedPattern, normalizedValues, ignore.case = TRUE) ]
+    }
+    else exact.matches
+}
+
+
+findMatches <- function(pattern, values)
+{
+    if (.CompletionEnv$settings[["fuzzy"]])
+        findFuzzyMatches(pattern, values)
+    else
+        findExactMatches(pattern, values)
+}
+
+fuzzyApropos <- function(what)
+{
+    stopifnot(is.character(what))
+    x <- character(0L)
+    for (i in seq_along(search())) {
+	li <- findFuzzyMatches(what, ls(pos = i, all.names = TRUE))
+	if (length(li)) { x <- c(x, li) }
+    }
+    findFuzzyMatches(what, x)
+}
+
 ## generic and built-in methods to generate completion after $
 
 .DollarNames <- function(x, pattern)
@@ -52,15 +137,18 @@
 
 .DollarNames.default <- function(x, pattern = "") {
     if (is.atomic(x) || is.symbol(x)) character()
-    else grep(pattern, names(x), value = TRUE)
+    else findMatches(pattern, names(x))
 }
 
 .DollarNames.list <- function(x, pattern = "") {
-    grep(pattern, names(x), value = TRUE)
+    findMatches(pattern, names(x))
 }
 
 .DollarNames.environment <- function(x, pattern = "") {
-    ls(x, all.names = TRUE, pattern = pattern)
+    if (!.CompletionEnv$settings[["fuzzy"]])
+        ls(x, all.names = TRUE, pattern = pattern) # more efficient
+    else
+        findMatches(pattern, ls(x, all.names = TRUE))
 }
 
 ## if (is.environment(object))
@@ -80,8 +168,9 @@
 
 ## modifies settings:
 
-rc.settings <- function(ops, ns, args, func, ipck, S3, data, help, argdb, files)
+rc.settings <- function(ops, ns, args, func, ipck, S3, data, help, argdb, fuzzy, quotes, files)
 {
+    if (length(match.call()) == 1) return(unlist(.CompletionEnv[["settings"]]))
     checkAndChange <- function(what, value)
     {
         if ((length(value) == 1L) &&
@@ -99,13 +188,13 @@ rc.settings <- function(ops, ns, args, func, ipck, S3, data, help, argdb, files)
     if (!missing(help))  checkAndChange( "help",  help)
     if (!missing(argdb)) checkAndChange("argdb", argdb)
     if (!missing(files)) checkAndChange("files", files)
+    if (!missing(quotes))checkAndChange("quotes", quotes)
+    if (!missing(fuzzy)) checkAndChange("fuzzy", fuzzy)
     invisible()
 }
 
 
 
-
-
 ## modifies options (adapted from similar functions in lattice):
 
 rc.getOption <- function(name)
@@ -126,7 +215,7 @@ rc.options <- function(...)
     nm <- names(new)
     if (is.null(nm)) return(old[unlist(new)])
 
-    isNamed <- nm != ""
+    isNamed <- nzchar(nm)
     if (any(!isNamed)) nm[!isNamed] <- unlist(new[!isNamed])
 
     ## so now everything has non-"" names, but only the isNamed ones
@@ -176,8 +265,11 @@ rc.status <- function()
 
 .guessTokenFromLine <-
     function(linebuffer = .CompletionEnv[["linebuffer"]],
-             end = .CompletionEnv[["end"]])
+             end = .CompletionEnv[["end"]],
+             update = TRUE)
+    ## update=TRUE changes 'start' and 'token', otherwise they are just returned
 {
+    linebuffer <- substr(linebuffer, 1L, end) # end is cursor, not necessarily line-end
     ## special rules apply when we are inside quotes (see fileCompletionPreferred() below)
     insideQuotes <- {
         lbss <- head.default(unlist(strsplit(linebuffer, "")), .CompletionEnv[["end"]])
@@ -187,36 +279,36 @@ rc.status <- function()
     start <-
         if (insideQuotes)
             ## set 'start' to the location of the last quote
-            suppressWarnings(gregexpr("['\"]",
-                                      substr(linebuffer, 1L, end),
+            suppressWarnings(gregexpr("['\"]", linebuffer,
                                       perl = TRUE))[[1L]]
-            ## suppressWarnings(gregexpr("[^\\.\\w:?$@[\\]\\\\/~ ]+", substr(linebuffer, 1L, end), perl = TRUE))[[1L]]
         else
             ##                    things that should not cause breaks
-            ##                           ______________
+            ##                           _____.^._____
+            ##                          /             \
             suppressWarnings(gregexpr("[^\\.\\w:?$@[\\]]+",
-                                      substr(linebuffer, 1L, end),
+                                      linebuffer,
                                       perl = TRUE))[[1L]]
     start <- ## 0-indexed
         if (all(start < 0L)) 0L
         else tail.default(start + attr(start, "match.length"), 1L) - 1L
-    .CompletionEnv[["start"]] <- start
-    .CompletionEnv[["token"]] <- substr(linebuffer, start + 1L, end)
-    .CompletionEnv[["token"]]
+    token <- substr(linebuffer, start + 1L, end)
+    if (update) {
+        .CompletionEnv[["start"]] <- start
+        .CompletionEnv[["token"]] <- token
+        .CompletionEnv[["token"]]
+    }
+    else list(start = start, token = token)
 }
 
 
 
 
-
-
 ## convert a string to something that escapes special regexp
 ## characters.  Doesn't have to be perfect, especially for characters
 ## that would cause breaks or be handled elsewhere.  All we really
 ## need is to handle ".", so that e.g. "heat." doesn't match
 ## "heatmap".
 
-
 makeRegexpSafe <- function(s)
 {
     ## the following can cause errors otherwise
@@ -243,6 +335,80 @@ makeRegexpSafe <- function(s)
 specialOps <- c("$", "@", ":::", "::", "?", "[", "[[")
 
 
+specialOpCompletionsHelper <- function(op, suffix, prefix)
+{
+    tryToEval <- function(s)
+    {
+	tryCatch(eval(parse(text = s), envir = .GlobalEnv), error = function(e)e)
+    }
+    switch(op,
+           "$" = {
+               if (.CompletionEnv$settings[["ops"]])
+               {
+                   object <- tryToEval(prefix)
+                   if (inherits(object, "error")) ## nothing else to do
+                       suffix
+                   else
+                   {
+                       ## ## suffix must match names(object) (or ls(object) for environments)
+                       .DollarNames(object, pattern = sprintf("^%s", makeRegexpSafe(suffix)))
+                   }
+               } else suffix
+           },
+           "@" = {
+               if (.CompletionEnv$settings[["ops"]])
+               {
+                   object <- tryToEval(prefix)
+                   if (inherits(object, "error")) ## nothing else to do
+                       suffix
+                   else
+                   {
+                       findMatches(sprintf("^%s", makeRegexpSafe(suffix)),
+                                   methods::slotNames(object))
+                   }
+               } else suffix
+           },
+           "::" = {
+               if (.CompletionEnv$settings[["ns"]])
+               {
+                   nse <- tryCatch(getNamespaceExports(prefix), error = function(e)e)
+                   if (inherits(nse, "error")) ## nothing else to do
+                       suffix
+                   else
+                   {
+                       findMatches(sprintf("^%s", makeRegexpSafe(suffix)),
+                                   nse)
+                   }
+               } else suffix
+           },
+           ":::" = {
+               if (.CompletionEnv$settings[["ns"]])
+               {
+                   ns <- tryCatch(getNamespace(prefix), error = function(e)e)
+                   if (inherits(ns, "error")) ## nothing else to do
+                       suffix
+                   else
+                   {
+                       if (!.CompletionEnv$settings[["fuzzy"]])
+                           ls(ns,
+                              all.names = TRUE,
+                              pattern = sprintf("^%s", makeRegexpSafe(suffix)))
+                       else
+                           findMatches(sprintf("^%s", makeRegexpSafe(suffix)),
+                                       ls(ns, all.names = TRUE))
+                   }
+               } else suffix
+           },
+           "[" = ,  # can't think of anything else to do
+           "[[" = {
+               comps <- normalCompletions(suffix)
+               if (length(comps)) comps
+               else suffix
+           })
+}
+
+
+
 
 specialOpLocs <- function(text)
 {
@@ -261,10 +427,10 @@ specialOpLocs <- function(text)
 
 
 
-## accessing the help system: should allow anything with an index entry
-## this just looks at packages on the search path.
+## Accessing the help system: should allow anything with an index entry.
+## This just looks at packages on the search path.
 
-matchAvailableTopics <- function(text)
+matchAvailableTopics <- function(prefix, text)
 {
     .readAliases <- function(path) {
         if(file.exists(f <- file.path(path, "help", "aliases.rds")))
@@ -276,25 +442,42 @@ matchAvailableTopics <- function(text)
         else character()
     }
     if (length(text) != 1L || text == "") return (character())
+    ## Update list of help topics if necessary
     pkgpaths <- searchpaths()[substr(search(), 1L, 8L) == "package:"]
-    aliases <- unique(unlist(lapply(pkgpaths, .readAliases)))
-    grep(sprintf("^%s", makeRegexpSafe(text)), aliases, value = TRUE)
+    if (!identical(basename(pkgpaths), .CompletionEnv[["attached_packages"]])) {
+        assign("attached_packages",
+               basename(pkgpaths),
+               envir = .CompletionEnv)
+        assign("help_topics",
+               unique(unlist(lapply(pkgpaths, .readAliases))),
+               envir = .CompletionEnv)
+    }
+    aliases <- .CompletionEnv[["help_topics"]]
+    ans <- findMatches(sprintf("^%s", makeRegexpSafe(text)), aliases)
+    if (nzchar(prefix)) {
+        ## FIXME: This is a little unsafe.  We are not protecting
+        ## prefix to make sure that we do not get any special
+        ## characters (like ? or + or *).  However, these are unlikely
+        ## in practice.
+        tmp <- grep(sprintf("-%s$", prefix), ans, value = TRUE)
+        if (length(tmp)) substring(tmp, 1, nchar(tmp) - nchar(prefix) - 1L)
+        else character(0)
+    }
+    else ans
 }
 
 
 
 ## this is for requests of the form ?suffix[TAB] or prefix?suffix[TAB]
 
-## can be improved when prefix is non-trivial, but that is rarely used
-## (on the other hand, can be useful for exploring S4 methods; but
-## usage of ? needs to be fixed in R first).  Anyway, that case is not
-## currently handled
 
-helpCompletions <- function(prefix, suffix)
+helpCompletions <- function(prefix = "", suffix)
 {
+    ## Do not attempt to complete ??<foo> (help.search) or ???<foo> (invalid)
+    if (prefix %in% c("?", "??")) return (character(0))
     nc <-
         if (.CompletionEnv$settings[["help"]])
-            matchAvailableTopics(suffix)
+            matchAvailableTopics(prefix, suffix)
         else
             normalCompletions(suffix, check.mode = FALSE)
     if (length(nc)) sprintf("%s?%s", prefix, nc)
@@ -335,83 +518,7 @@ specialCompletions <- function(text, spl)
     ## ( breaks words, so prefix should not involve function calls,
     ## and thus, hopefully no side-effects.
 
-    tryToEval <- function(s)
-    {
-        try(eval(parse(text = s), envir = .GlobalEnv), silent = TRUE)
-    }
-
-    comps <-
-        switch(op,
-               "$" = {
-                   if (.CompletionEnv$settings[["ops"]])
-                   {
-                       object <- tryToEval(prefix)
-                       if (inherits(object, "try-error")) ## nothing else to do
-                           suffix
-                       else
-                       {
-                           ## ## suffix must match names(object) (or ls(object) for environments)
-                           .DollarNames(object, pattern = sprintf("^%s", makeRegexpSafe(suffix)))
-                           ## if (is.environment(object))
-                           ## {
-                           ##     ls(object,
-                           ##        all.names = TRUE,
-                           ##        pattern = sprintf("^%s", makeRegexpSafe(suffix)))
-                           ## }
-                           ## else
-                           ## {
-                           ##     grep(sprintf("^%s", makeRegexpSafe(suffix)),
-                           ##          names(object), value = TRUE)
-                           ## }
-                       }
-                   } else suffix
-               },
-               "@" = {
-                   if (.CompletionEnv$settings[["ops"]])
-                   {
-                       object <- tryToEval(prefix)
-                       if (inherits(object, "try-error")) ## nothing else to do
-                           suffix
-                       else
-                       {
-                           grep(sprintf("^%s", makeRegexpSafe(suffix)),
-                                methods::slotNames(object), value = TRUE)
-                       }
-                   } else suffix
-               },
-               "::" = {
-                   if (.CompletionEnv$settings[["ns"]])
-                   {
-                       nse <- try(getNamespaceExports(prefix), silent = TRUE)
-                       if (inherits(nse, "try-error")) ## nothing else to do
-                           suffix
-                       else
-                       {
-                           grep(sprintf("^%s", makeRegexpSafe(suffix)),
-                                nse, value = TRUE)
-                       }
-                   } else suffix
-               },
-               ":::" = {
-                   if (.CompletionEnv$settings[["ns"]])
-                   {
-                       ns <- try(getNamespace(prefix), silent = TRUE)
-                       if (inherits(ns, "try-error")) ## nothing else to do
-                           suffix
-                       else
-                       {
-                           ls(ns,
-                              all.names = TRUE,
-                              pattern = sprintf("^%s", makeRegexpSafe(suffix)))
-                       }
-                   } else suffix
-               },
-               "[" = ,  # can't think of anything else to do
-               "[[" = {
-                   comps <- normalCompletions(suffix)
-                   if (length(comps)) comps
-                   else suffix
-               })
+    comps <- specialOpCompletionsHelper(op, suffix, prefix)
     if (length(comps) == 0L) comps <- ""
     sprintf("%s%s%s", prefix, op, comps)
 }
@@ -427,31 +534,49 @@ specialCompletions <- function(text, spl)
 
 keywordCompletions <- function(text)
 {
-    grep(sprintf("^%s", makeRegexpSafe(text)),
-         c("NULL", "NA", "TRUE", "FALSE", "Inf", "NaN",
-           "NA_integer_", "NA_real_", "NA_character_", "NA_complex_",
-           "repeat ", "in ", "next ", "break ", "else "),
-         value = TRUE)
+    ## FIXME: Will not allow fuzzy completions, as this adds too much
+    ## noise in normalCompletions.  Should revisit later once we
+    ## figure out a way to suppress fuzzy matching if there is at
+    ## least one exact match.
+    findExactMatches(sprintf("^%s", makeRegexpSafe(text)),
+                     c("NULL", "NA", "TRUE", "FALSE", "Inf", "NaN",
+                       "NA_integer_", "NA_real_", "NA_character_", "NA_complex_",
+                       "repeat ", "in ", "next ", "break ", "else "))
 }
 
 
 
 
 ## 'package' environments in the search path.  These will be completed
-## with a :: IIRC, that works even if there's no namespace, but I
-## haven't actually checked.
-
+## with a :: (Use of this is function is replaced by
+## loadedPackageCompletions below, which also completes packages
+## loaded, but not necessarily attached).
 
 
 attachedPackageCompletions <- function(text, add = rc.getOption("package.suffix"))
 {
+    ## FIXME: Will not allow fuzzy completions. See comment in keywordCompletions() above
     if (.CompletionEnv$settings[["ns"]])
     {
         s <- grep("^package", search(), value = TRUE)
         comps <-
-            grep(sprintf("^%s", makeRegexpSafe(text)),
-                 substr(s, 9L, 1000000L),
-                 value = TRUE)
+            findExactMatches(sprintf("^%s", makeRegexpSafe(text)),
+                             substr(s, 9L, 1000000L))
+        if (length(comps) && !is.null(add))
+            sprintf("%s%s", comps, add)
+        else
+            comps
+    }
+    else character()
+}
+
+loadedPackageCompletions <- function(text, add = rc.getOption("package.suffix"))
+{
+    ## FIXME: Will not allow fuzzy completions. See comment in keywordCompletions() above
+    if (.CompletionEnv$settings[["ns"]])
+    {
+        s <- loadedNamespaces()
+        comps <- findExactMatches(sprintf("^%s", makeRegexpSafe(text)), s)
         if (length(comps) && !is.null(add))
             sprintf("%s%s", comps, add)
         else
@@ -464,7 +589,7 @@ attachedPackageCompletions <- function(text, add = rc.getOption("package.suffix"
 
 ## this provides the most basic completion, looking for completions in
 ## the search path using apropos, plus keywords.  Plus completion on
-## attached packages if settings$ns == TRUE
+## attached/loaded packages if settings$ns == TRUE
 
 
 normalCompletions <-
@@ -475,7 +600,11 @@ normalCompletions <-
     if (text == "") character() ## too many otherwise
     else
     {
-        comps <- apropos(sprintf("^%s", makeRegexpSafe(text)), ignore.case = FALSE)
+        comps <-
+            if (.CompletionEnv$settings[["fuzzy"]])
+                fuzzyApropos(sprintf("^%s", makeRegexpSafe(text)))
+            else
+                apropos(sprintf("^%s", makeRegexpSafe(text)), ignore.case = FALSE)
         if (.CompletionEnv$settings[["func"]] && check.mode && !is.null(add.fun))
         {
             which.function <- sapply(comps, function(s) exists(s, mode = "function"))
@@ -484,7 +613,7 @@ normalCompletions <-
                     sprintf("%s%s", comps[which.function], add.fun)
             ##sprintf("\033[31m%s\033[0m%s", comps[which.function], add.fun)
         }
-        c(comps, keywordCompletions(text), attachedPackageCompletions(text))
+        c(comps, keywordCompletions(text), loadedPackageCompletions(text))
     }
 }
 
@@ -553,7 +682,7 @@ inFunction <-
 
     temp <-
         data.frame(i = c(parens[["("]], parens[[")"]]),
-                   c = rep(c(1, -1), sapply(parens, length)))
+                   c = rep(c(1, -1), lengths(parens)))
     if (nrow(temp) == 0) return(character())
     temp <- temp[order(-temp$i), , drop = FALSE] ## order backwards
     wp <- which(cumsum(temp$c) > 0)
@@ -586,7 +715,7 @@ inFunction <-
         else ## guess function name
         {
             possible <- suppressWarnings(strsplit(prefix, breakRE, perl = TRUE))[[1L]]
-            possible <- possible[possible != ""]
+            possible <- possible[nzchar(possible)]
             if (length(possible)) return(tail.default(possible, 1))
             else return(character())
         }
@@ -630,8 +759,8 @@ specialFunctionArgs <- function(fun, text)
            require = {
                if (.CompletionEnv$settings[["ipck"]])
                {
-                   grep(sprintf("^%s", makeRegexpSafe(text)),
-                        rownames(installed.packages()), value = TRUE)
+                   findMatches(sprintf("^%s", makeRegexpSafe(text)),
+                               rownames(installed.packages()))
                }
                else character()
            },
@@ -639,8 +768,8 @@ specialFunctionArgs <- function(fun, text)
            data = {
                if (.CompletionEnv$settings[["data"]])
                {
-                   grep(sprintf("^%s", makeRegexpSafe(text)),
-                        data()$results[, "Item"], value = TRUE)
+                   findMatches(sprintf("^%s", makeRegexpSafe(text)),
+                               data()$results[, "Item"])
                }
                else character()
            },
@@ -665,12 +794,9 @@ functionArgs <-
               tryCatch(methods(fun),
                        warning = function(w) {},
                        error = function(e) {}))
-    if (S4methods)
-        warning("Can't handle S4 methods yet")
+    if (S4methods) warning("cannot handle S4 methods yet")
     allArgs <- unique(unlist(lapply(fun, argNames)))
-    ans <-
-        grep(sprintf("^%s", makeRegexpSafe(text)),
-             allArgs, value = TRUE)
+    ans <- findMatches(sprintf("^%s", makeRegexpSafe(text)), allArgs)
     if (length(ans) && !is.null(add.args))
         ans <- sprintf("%s%s", ans, add.args)
     c(specialFunArgs, ans)
@@ -708,10 +834,15 @@ functionArgs <-
 ## number of quotes between the cursor and the beginning of the line
 ## is an odd number.
 
+## FIXME: should include backtick (`)? May be useful, but needs more
+## thought; e.g., should imply not-filename, but rather variable
+## names.  Must cooperate with the if (isInsideQuotes()) branch in
+## .completeToken().
+
 isInsideQuotes <-
 fileCompletionPreferred <- function()
 {
-    ((st <- .CompletionEnv[["start"]]) > 0 && {
+    (.CompletionEnv[["start"]] > 0 && {
 
         ## yes if the number of quote signs to the left is odd
         linebuffer <- .CompletionEnv[["linebuffer"]]
@@ -723,9 +854,9 @@ fileCompletionPreferred <- function()
 }
 
 
-## File name completion, used if settings$files == TRUE.  Front ends
+## File name completion, used if settings$quotes == TRUE.  Front ends
 ## that can do filename completion themselves should probably not use
-## this as they will do a better job.
+## this if they can do a better job.
 
 correctFilenameToken <- function()
 {
@@ -773,6 +904,18 @@ fileCompletions <- function(token)
     pfilename.expanded <- path.expand(pfilename)
     comps <- Sys.glob(sprintf("%s*", pfilename.expanded), dirmark = TRUE)
 
+    ## If there is only one completion (and it's a directory), also
+    ## include files inside in list of completions.  This is not
+    ## particularly useful, but without this, readline tends to add an
+    ## end-quote (if sole completion) which is irritating if one is
+    ## actually looking for something inside the directory.  Note that
+    ## we don't actually test to see if it's a directory, because if
+    ## it is not, list.files() will simply return character(0).
+    if (length(comps) == 1 && substring(comps, nchar(comps), nchar(comps)) == "/") {
+        filesInside <- list.files(comps, all.files = TRUE, full.names = FALSE, no.. = TRUE)
+        if (length(filesInside)) comps <- c(comps, file.path(comps, filesInside))
+    }
+
     ## for things that only extend beyond the cursor, need to
     ## 'unexpand' path
     if (pfilename.expanded != pfilename)
@@ -782,7 +925,6 @@ fileCompletions <- function(token)
     ## need to delete extra part
     if (pfilename != token)
         comps <- substring(comps, nchar(pfilename) - nchar(token) + 1L, 1000L)
-
     comps
 }
 
@@ -796,6 +938,10 @@ fileCompletions <- function(token)
 
 .completeToken <- function()
 {
+    ## Allow override by user-specified function
+    custom.completer <- rc.getOption("custom.completer")
+    if (is.function(custom.completer))
+        return (custom.completer(.CompletionEnv))
     text <- .CompletionEnv[["token"]]
     if (isInsideQuotes())
     {
@@ -810,40 +956,116 @@ fileCompletions <- function(token)
         ## We assume that whoever determines our token boundaries
         ## considers quote signs as a breaking symbol.
 
-        st <- .CompletionEnv[["start"]]
-        probablyNotFilename <-
-            (st > 2L &&
-             (substr(.CompletionEnv[["linebuffer"]], st-1L, st-1L) %in% c("[", ":", "$")))
 
-
-        ## If the 'files' setting is FALSE, we will make no attempt to
+        ## If the 'quotes' setting is FALSE, we will make no attempt to
         ## do filename completion (this is likely to happen with
         ## front-ends that are capable of doing their own file name
         ## completion; such front-ends can fall back to their native
         ## file completion when rc.status("fileName") is TRUE.
 
-        if (.CompletionEnv$settings[["files"]])
+        if (.CompletionEnv$settings[["quotes"]])
         {
 
-            ## we bail out if probablyNotFilename == TRUE.  If we
-            ## wanted to be really fancy, we could try to detect where
-            ## we are and act accordingly, e.g. if it's
-            ## foo[["bar<TAB>, pretend we are completing foo$bar, etc.
-            ## This is not that hard, we just need to use
-            ## .guessTokenFromLine(linebuffer, end = st - 1) to get the
-            ## 'foo[[' part.
+            ## ## This was used to make a guess whether we are in
+            ## ## special situations like ::, ?, [, etc.  But from R
+            ## ## 3.0.0 we re-evaluate the token based from the
+            ## ## begin-quote, so this is postponed.  This part can be
+            ## ## deleted once this is stable enough.
+            ## st <- .CompletionEnv[["start"]]
+            ## probablyNotFilename <-
+            ##     ((st > 2L &&
+            ##       ((prequote <- substr(.CompletionEnv[["linebuffer"]], st-1L, st-1L)) %in% c("?", "[", ":", "$"))) ||
+            ##      (st == 2L &&
+            ##       ((prequote <- substr(.CompletionEnv[["linebuffer"]], st-1L, st-1L)) %in% c("?")))
+            ##      )
 
-            .CompletionEnv[["comps"]] <-
-                if (probablyNotFilename) character()
-                else fileCompletions(text)
+            ## FIXME|TODO: readline (and maybe other backends) will
+            ## usually use a fixed set of breakpoints to detect
+            ## tokens.  If we are handling quotes ourselves, the more
+            ## likely correct token is everything from the last
+            ## unclosed quote onwards (which may include spaces,
+            ## punctuations, etc. that would normally cause breaks).
+            ## We already do this when we guess the token ourselves
+            ## (e.g., for Windows) (and also in the fileCompletions()
+            ## call below using correctFilenameToken()), and can
+            ## re-use that here.  The problem is that for other
+            ## backends a token may already have been determined, and
+            ## that's what we will need to use.  We can still fake it
+            ## by using the correct token but substracting the extra
+            ## part when providing completions, but that will need
+            ## some work.
+
+            ## Related to that: if we implement that, should also
+            ## check before for '<type>?' and move to help completion
+            ## if so.
+
+### str(correctFilenameToken())
+### str(.guessTokenFromLine(update = FALSE))
+
+            ## TODO: For extra credit, we could also allow for
+            ## spaces like in 'package ? grid', but will leave
+            ## that for the future (maybe some regexp magic will
+            ## make this simple)
+
+            fullToken <- .guessTokenFromLine(update = FALSE)
+            probablyHelp <- (fullToken$start >= 2L &&
+                             ((substr(.CompletionEnv[["linebuffer"]],
+                                      fullToken$start-1L,
+                                      fullToken$start-1L)) == "?"))
+            if (probablyHelp) {
+                fullToken$prefix <- .guessTokenFromLine(end = fullToken$start - 2, update = FALSE)$token
+            }
+            probablyName <- ((fullToken$start > 2L &&
+                              ((substr(.CompletionEnv[["linebuffer"]],
+                                       fullToken$start-1L,
+                                       fullToken$start-1L)) == "$"))
+                             ||
+                             (fullToken$start > 3L &&
+                              ((substr(.CompletionEnv[["linebuffer"]],
+                                       fullToken$start-2L,
+                                       fullToken$start-1L)) == "[[")))
+            probablyNamespace <- (fullToken$start > 3L &&
+                                  ((substr(.CompletionEnv[["linebuffer"]],
+                                           fullToken$start-2L,
+                                           fullToken$start-1L)) %in% c("::")))
+            ## in anticipation that we will handle this eventually:
+##             probablyBacktick <- (fullToken$start >= 1L &&
+##                                  ((substr(.CompletionEnv[["linebuffer"]],
+##                                           fullToken$start,
+##                                           fullToken$start)) %in% c("`")))
+
+            probablySpecial <- probablyHelp || probablyName || probablyNamespace
+
+            ## str(list(probablyHelp = probablyHelp,
+            ##          probablyName = probablyName,
+            ##          probablyNamespace = probablyNamespace,
+            ##          probablyBacktick = probablyBacktick,
+            ##          probablySpecial = probablySpecial))
+
+            ## For now, we only handle probablyHelp, and just decline
+            ## to do filename completion if any of the other special
+            ## situations are detected (but don't try to complete).
+
+            tentativeCompletions <-
+                if (probablyHelp) {
+                    substring(helpCompletions(fullToken$prefix, fullToken$token),
+                              2L + nchar(fullToken$prefix), 1000L)    # drop initial "prefix + ?"
+                }
+                else if (!probablySpecial)
+                    fileCompletions(fullToken$token) # FIXME: but not if probablyBacktick
             .setFileComp(FALSE)
+            ## str(c(fullToken, list(comps = tentativeCompletions)))
+            ## Adjust for self-computed token
+            .CompletionEnv[["comps"]] <-
+                substring(tentativeCompletions,
+                          1L + nchar(fullToken$token) - nchar(text),
+                          1000L)
         }
         else
         {
             .CompletionEnv[["comps"]] <- character()
             .setFileComp(TRUE)
         }
-
     }
     else
     {
@@ -911,7 +1133,7 @@ fileCompletions <- function(token)
         {
             comps <- paste0(prefix, comps)
         }
-        comps <- c(comps, fargComps)
+        comps <- c(fargComps, comps)
         .CompletionEnv[["comps"]] <- comps
     }
 }
@@ -1114,18 +1336,24 @@ assign("settings",
        list(ops = TRUE, ns = TRUE,
             args = TRUE, func = FALSE,
             ipck = FALSE, S3 = TRUE, data = TRUE,
-            help = TRUE, argdb = TRUE,
-            files = .Platform$OS.type == "windows"),
+            help = TRUE, argdb = TRUE, fuzzy = FALSE,
+            files = TRUE, # FIXME: deprecate in favour of quotes
+            quotes = TRUE),
        env = .CompletionEnv)
 
-
 assign("options",
        list(package.suffix = "::",
             funarg.suffix = "=",
             function.suffix = "("),
        env = .CompletionEnv)
 
+## These keeps track of attached packages and available help topics.
+## Needs updating only when packages are attached.
+assign("attached_packages", character(0), env = .CompletionEnv)
+assign("help_topics", character(0), env = .CompletionEnv)
+
+
 .FunArgEnv <- new.env(hash = TRUE, parent = emptyenv())
 
-.initialize.argdb() ## see argdb.R
+.initialize.argdb()
 
