@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2001-2010  The R Core Team
+ *  Copyright (C) 2001-2014  The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -42,34 +42,43 @@ static Rboolean neWithNaN(double x, double y, ne_strictness_type str);
 /* .Internal(identical(..)) */
 SEXP attribute_hidden do_identical(SEXP call, SEXP op, SEXP args, SEXP env)
 {
-    int num_eq = 1, single_NA = 1, attr_as_set = 1, ignore_bytecode = 1, nargs = length(args), flags;
-    /* avoid problems with earlier (and future) versions captured in S4 methods */
-    /* checkArity(op, args); */
+    int num_eq = 1, single_NA = 1, attr_as_set = 1, ignore_bytecode = 1,
+	ignore_env = 0, nargs = length(args), flags;
+    /* avoid problems with earlier (and future) versions captured in S4
+       methods: but this should be fixed where it is caused, in
+       'methods'!
+
+       checkArity(op, args); */
     if (nargs < 5)
 	error("%d arguments passed to .Internal(%s) which requires %d",
 	      length(args), PRIMNAME(op), PRIMARITY(op));
 
-    if (nargs >= 5) {
-	num_eq      = asLogical(CADDR(args));
-	single_NA   = asLogical(CADDDR(args));
-	attr_as_set = asLogical(CAD4R(args));
-    }
-    if (nargs >= 6) 
-	ignore_bytecode = asLogical(CAD4R(CDR(args)));
+    SEXP x = CAR(args); args = CDR(args);
+    SEXP y = CAR(args); args = CDR(args);
+    num_eq = asLogical(CAR(args)); args = CDR(args);
+    single_NA = asLogical(CAR(args)); args = CDR(args);
+    attr_as_set = asLogical(CAR(args)); args = CDR(args);
+    if (nargs >= 6)
+	ignore_bytecode = asLogical(CAR(args));
+    if (nargs >= 7)
+	ignore_env = asLogical(CADR(args));
 
-    if(num_eq      == NA_LOGICAL) error(_("invalid '%s' value"), "num.eq");
-    if(single_NA   == NA_LOGICAL) error(_("invalid '%s' value"), "single.NA");
+    if(num_eq == NA_LOGICAL) error(_("invalid '%s' value"), "num.eq");
+    if(single_NA == NA_LOGICAL) error(_("invalid '%s' value"), "single.NA");
     if(attr_as_set == NA_LOGICAL) error(_("invalid '%s' value"), "attrib.as.set");
     if(ignore_bytecode == NA_LOGICAL) error(_("invalid '%s' value"), "ignore.bytecode");
-    
-    flags = (num_eq ? 0 : 1) + (single_NA ? 0 : 2) + (attr_as_set ? 0 : 4) + (ignore_bytecode ? 0 : 8); 
-    return ScalarLogical(R_compute_identical(CAR(args), CADR(args), flags));
+    if(ignore_env == NA_LOGICAL) error(_("invalid '%s' value"), "ignore.environment");
+
+    flags = (num_eq ? 0 : 1) + (single_NA ? 0 : 2) + (attr_as_set ? 0 : 4) +
+	(ignore_bytecode ? 0 : 8) + (ignore_env ? 0 : 16);
+    return ScalarLogical(R_compute_identical(x, y, flags));
 }
 
 #define NUM_EQ 		(!(flags & 1))
 #define SINGLE_NA       (!(flags & 2))
 #define ATTR_AS_SET     (!(flags & 4))
 #define IGNORE_BYTECODE (!(flags & 8))
+#define IGNORE_ENV      (!(flags & 16))
 
 /* do the two objects compute as identical?
    Also used in unique.c */
@@ -79,9 +88,9 @@ R_compute_identical(SEXP x, SEXP y, int flags)
     SEXP ax, ay, atrx, atry;
     if(x == y) /* same pointer */
 	return TRUE;
-    if(TYPEOF(x) != TYPEOF(y))
-	return FALSE;
-    if(OBJECT(x) != OBJECT(y))
+    if(TYPEOF(x) != TYPEOF(y) ||
+       OBJECT(x) != OBJECT(y) ||
+       IS_S4_OBJECT(x) != IS_S4_OBJECT(y))
 	return FALSE;
 
     /* Skip attribute checks for CHARSXP
@@ -141,35 +150,35 @@ R_compute_identical(SEXP x, SEXP y, int flags)
     case NILSXP:
 	return TRUE;
     case LGLSXP:
-	if (length(x) != length(y)) return FALSE;
+	if (xlength(x) != xlength(y)) return FALSE;
 	/* Use memcmp (which is ISO C90) to speed up the comparison */
 	return memcmp((void *)LOGICAL(x), (void *)LOGICAL(y),
-		      length(x) * sizeof(int)) == 0 ? TRUE : FALSE;
+		      xlength(x) * sizeof(int)) == 0 ? TRUE : FALSE;
     case INTSXP:
-	if (length(x) != length(y)) return FALSE;
+	if (xlength(x) != xlength(y)) return FALSE;
 	/* Use memcmp (which is ISO C90) to speed up the comparison */
 	return memcmp((void *)INTEGER(x), (void *)INTEGER(y),
-		      length(x) * sizeof(int)) == 0 ? TRUE : FALSE;
+		      xlength(x) * sizeof(int)) == 0 ? TRUE : FALSE;
     case REALSXP:
     {
-	int n = length(x);
-	if(n != length(y)) return FALSE;
+	R_xlen_t n = xlength(x);
+	if(n != xlength(y)) return FALSE;
 	else {
 	    double *xp = REAL(x), *yp = REAL(y);
-	    int i, ne_strict = NUM_EQ | (SINGLE_NA << 1);
-	    for(i = 0; i < n; i++)
+	    int ne_strict = NUM_EQ | (SINGLE_NA << 1);
+	    for(R_xlen_t i = 0; i < n; i++)
 		if(neWithNaN(xp[i], yp[i], ne_strict)) return FALSE;
 	}
 	return TRUE;
     }
     case CPLXSXP:
     {
-	int n = length(x);
-	if(n != length(y)) return FALSE;
+	R_xlen_t n = xlength(x);
+	if(n != xlength(y)) return FALSE;
 	else {
 	    Rcomplex *xp = COMPLEX(x), *yp = COMPLEX(y);
-	    int i, ne_strict = NUM_EQ | (SINGLE_NA << 1);
-	    for(i = 0; i < n; i++)
+	    int ne_strict = NUM_EQ | (SINGLE_NA << 1);
+	    for(R_xlen_t i = 0; i < n; i++)
 		if(neWithNaN(xp[i].r, yp[i].r, ne_strict) ||
 		   neWithNaN(xp[i].i, yp[i].i, ne_strict))
 		    return FALSE;
@@ -178,8 +187,8 @@ R_compute_identical(SEXP x, SEXP y, int flags)
     }
     case STRSXP:
     {
-	int i, n = length(x);
-	if(n != length(y)) return FALSE;
+	R_xlen_t i, n = xlength(x);
+	if(n != xlength(y)) return FALSE;
 	for(i = 0; i < n; i++) {
 	    /* This special-casing for NAs is not needed */
 	    Rboolean na1 = (STRING_ELT(x, i) == NA_STRING),
@@ -198,8 +207,8 @@ R_compute_identical(SEXP x, SEXP y, int flags)
     case VECSXP:
     case EXPRSXP:
     {
-	int i, n = length(x);
-	if(n != length(y)) return FALSE;
+	R_xlen_t i, n = xlength(x);
+	if(n != xlength(y)) return FALSE;
 	for(i = 0; i < n; i++)
 	    if(!R_compute_identical(VECTOR_ELT(x, i),VECTOR_ELT(y, i), flags))
 		return FALSE;
@@ -223,7 +232,7 @@ R_compute_identical(SEXP x, SEXP y, int flags)
     case CLOSXP:
 	return(R_compute_identical(FORMALS(x), FORMALS(y), flags) &&
 	       R_compute_identical(BODY_EXPR(x), BODY_EXPR(y), flags) &&
-	       (CLOENV(x) == CLOENV(y) ? TRUE : FALSE) &&
+	       (IGNORE_ENV || CLOENV(x) == CLOENV(y) ? TRUE : FALSE) &&
 	       (IGNORE_BYTECODE || R_compute_identical(BODY(x), BODY(y), flags))
 	       );
     case SPECIALSXP:
@@ -237,10 +246,10 @@ R_compute_identical(SEXP x, SEXP y, int flags)
     case EXTPTRSXP:
 	return (EXTPTR_PTR(x) == EXTPTR_PTR(y) ? TRUE : FALSE);
     case RAWSXP:
-	if (length(x) != length(y)) return FALSE;
+	if (xlength(x) != xlength(y)) return FALSE;
 	/* Use memcmp (which is ISO C90) to speed up the comparison */
 	return memcmp((void *)RAW(x), (void *)RAW(y),
-		      length(x) * sizeof(Rbyte)) == 0 ? TRUE : FALSE;
+		      xlength(x) * sizeof(Rbyte)) == 0 ? TRUE : FALSE;
 
 /*  case PROMSXP: args are evaluated, so will not be seen */
 	/* test for equality of the substituted expression -- or should
