@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2001-4   The R Development Core Team.
+ *  Copyright (C) 2001-2015   The R Core Team.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,6 +22,8 @@
 #endif
 
 #include <Defn.h>
+#include <Internal.h>
+
 #include <Rconnections.h>
 #include <Rdynpriv.h>
 #include <R_ext/R-ftp-http.h>
@@ -31,7 +33,7 @@ static R_InternetRoutines routines, *ptr = &routines;
 
 
 /*
-SEXP do_download(SEXP call, SEXP op, SEXP args, SEXP env);
+SEXP Rdownload(SEXP args);
 Rconnection R_newurl(char *description, char *mode);
 Rconnection R_newsock(char *host, int port, int server, char *mode, int timeout);
 
@@ -71,19 +73,10 @@ R_setInternetRoutines(R_InternetRoutines *routines)
     return(tmp);
 }
 
-#ifdef Win32
-extern Rboolean UseInternet2;
-#endif
-
 static void internet_Init(void)
 {
     int res;
-#ifdef Win32
-    res = UseInternet2 ? R_moduleCdynload("internet2", 1, 1) :
-	R_moduleCdynload("internet", 1, 1);
-#else
     res = R_moduleCdynload("internet", 1, 1);
-#endif
     initialized = -1;
     if(!res) return;
     if(!ptr->download)
@@ -92,12 +85,11 @@ static void internet_Init(void)
     return;
 }
 
-
-SEXP attribute_hidden do_download(SEXP call, SEXP op, SEXP args, SEXP env)
+SEXP Rdownload(SEXP args)
 {
     if(!initialized) internet_Init();
     if(initialized > 0)
-	return (*ptr->download)(call, op, args, env);
+	return (*ptr->download)(args);
     else {
 	error(_("internet routines cannot be loaded"));
 	return R_NilValue;
@@ -119,23 +111,19 @@ SEXP attribute_hidden do_setInternet2(SEXP call, SEXP op, SEXP args, SEXP env)
     
     if (newUseInternet2 != NA_LOGICAL) {
     	R_Visible = FALSE;
-    	if (newUseInternet2 != UseInternet2) {
-    	    if (initialized) warning(_("internet routines were already initialized"));
-    	    UseInternet2 = newUseInternet2;
-    	    initialized = 0;
-    	}
+	UseInternet2 = newUseInternet2;
     }
     UNPROTECT(1);
     return retval;
 }
 #endif
 
-Rconnection attribute_hidden R_newurl(const char *description,
-				      const char * const mode)
+Rconnection attribute_hidden 
+R_newurl(const char *description, const char * const mode, int type)
 {
     if(!initialized) internet_Init();
     if(initialized > 0)
-	return (*ptr->newurl)(description, mode);
+	return (*ptr->newurl)(description, mode, type);
     else {
 	error(_("internet routines cannot be loaded"));
 	return (Rconnection)0;
@@ -217,8 +205,7 @@ void  R_FTPClose(void *ctx)
 	error(_("internet routines cannot be loaded"));
 }
 
-attribute_hidden
-int   R_HTTPDCreate(const char *ip, int port)
+int extR_HTTPDCreate(const char *ip, int port)
 {
     if(!initialized) internet_Init();
     if(initialized > 0)
@@ -228,8 +215,7 @@ int   R_HTTPDCreate(const char *ip, int port)
     return -1;
 }
 
-attribute_hidden
-void R_HTTPDStop(void)
+void extR_HTTPDStop(void)
 {
     if(!initialized) internet_Init();
     if(initialized > 0)
@@ -238,87 +224,98 @@ void R_HTTPDStop(void)
 	error(_("internet routines cannot be loaded"));
 }
 
-SEXP attribute_hidden do_startHTTPD(SEXP call, SEXP op, SEXP args, SEXP env) 
-{
-    const char *ip = 0;
-    SEXP sIP, sPort;
-    checkArity(op, args);
-    sIP = CAR(args);
-    sPort = CADR(args);
-    if (sIP != R_NilValue && (TYPEOF(sIP) != STRSXP || LENGTH(sIP) != 1))
-	error(_("invalid bind address specification"));
-    if (sIP != R_NilValue)
-	ip = CHAR(STRING_ELT(sIP, 0));
-    return ScalarInteger(R_HTTPDCreate(ip, asInteger(sPort)));
-}
 
-SEXP attribute_hidden do_stopHTTPD(SEXP call, SEXP op, SEXP args, SEXP env) 
-{
-    checkArity(op, args);
-    R_HTTPDStop();
-    return R_NilValue;
-}
-
-attribute_hidden
-void Rsockopen(int *port)
+SEXP Rsockconnect(SEXP sport, SEXP shost)
 {
+    if (length(sport) != 1) error("invalid 'socket' argument");
+    int port = asInteger(sport);
+    char *host[1];
+    host[0] = (char *) translateChar(STRING_ELT(shost, 0));
     if(!initialized) internet_Init();
     if(initialized > 0)
-	(*ptr->sockopen)(port);
+	(*ptr->sockconnect)(&port, host);
     else
 	error(_("socket routines cannot be loaded"));
+    return ScalarInteger(port); // The socket number
 }
 
-attribute_hidden
-void Rsocklisten(int *sockp, char **buf, int *len)
+SEXP Rsockread(SEXP ssock, SEXP smaxlen)
 {
+    if (length(ssock) != 1) error("invalid 'socket' argument");
+    int sock = asInteger(ssock), maxlen = asInteger(smaxlen);
+    char buf[maxlen+1], *abuf[1];
+    abuf[0] = buf;
     if(!initialized) internet_Init();
     if(initialized > 0)
-	(*ptr->socklisten)(sockp, buf, len);
+	(*ptr->sockread)(&sock, abuf, &maxlen);
     else
 	error(_("socket routines cannot be loaded"));
+    SEXP ans = PROTECT(allocVector(STRSXP, 1));
+    SET_STRING_ELT(ans, 0, mkCharLen(buf, maxlen));
+    UNPROTECT(1);
+    return ans;
+		       
 }
 
-attribute_hidden
-void Rsockconnect(int *port, char **host)
+SEXP Rsockclose(SEXP ssock)
 {
+    if (length(ssock) != 1) error("invalid 'socket' argument");
+    int sock = asInteger(ssock);
     if(!initialized) internet_Init();
     if(initialized > 0)
-	(*ptr->sockconnect)(port, host);
+	(*ptr->sockclose)(&sock);
     else
 	error(_("socket routines cannot be loaded"));
+    return ScalarLogical(sock);
 }
 
-attribute_hidden
-void Rsockclose(int *sockp)
+SEXP Rsockopen(SEXP sport)
 {
+    if (length(sport) != 1) error("invalid 'port' argument");
+    int port = asInteger(sport);
     if(!initialized) internet_Init();
     if(initialized > 0)
-	(*ptr->sockclose)(sockp);
+	(*ptr->sockopen)(&port);
     else
 	error(_("socket routines cannot be loaded"));
+    return ScalarInteger(port); // The socket number
 }
 
-attribute_hidden
-void Rsockread(int *sockp, char **buf, int *maxlen)
+SEXP Rsocklisten(SEXP ssock)
 {
+    if (length(ssock) != 1) error("invalid 'socket' argument");
+    int sock = asInteger(ssock), len = 256;
+    char buf[257], *abuf[1];
+    abuf[0] = buf;
     if(!initialized) internet_Init();
     if(initialized > 0)
-	(*ptr->sockread)(sockp, buf, maxlen);
+	(*ptr->socklisten)(&sock, abuf, &len);
     else
 	error(_("socket routines cannot be loaded"));
+    SEXP ans = PROTECT(ScalarInteger(sock)); // The socket being listened on
+    SEXP host = PROTECT(allocVector(STRSXP, 1));
+    SET_STRING_ELT(host, 0, mkChar(buf));
+    setAttrib(ans, install("host"), host);
+    UNPROTECT(2);
+    return ans;
 }
 
-attribute_hidden
-void Rsockwrite(int *sockp, char **buf, int *start, int *end, int *len)
+SEXP Rsockwrite(SEXP ssock, SEXP sstring)
 {
+    if (length(ssock) != 1) error("invalid 'socket' argument");
+    int sock = asInteger(ssock), start = 0, end, len;
+    char *buf = (char *) translateChar(STRING_ELT(sstring, 0)), *abuf[1];
+    end = len = (int) strlen(buf);
+    abuf[0] = buf;
     if(!initialized) internet_Init();
     if(initialized > 0)
-	(*ptr->sockwrite)(sockp, buf, start, end, len);
+	(*ptr->sockwrite)(&sock, abuf, &start, &end, &len);
     else
 	error(_("socket routines cannot be loaded"));
+    return ScalarInteger(len);
 }
 
+
 attribute_hidden
 int Rsockselect(int nsock, int *insockfd, int *ready, int *write,
 		double timeout)
@@ -331,3 +328,50 @@ int Rsockselect(int nsock, int *insockfd, int *ready, int *write,
 	return 0;
     }
 }
+
+SEXP attribute_hidden do_curlVersion(SEXP call, SEXP op, SEXP args, SEXP rho)
+{
+    if(!initialized) internet_Init();
+    if(initialized > 0)
+	return (*ptr->curlVersion)(call, op, args, rho);
+    else {
+	error(_("internet routines cannot be loaded"));
+	return R_NilValue;
+    }
+}
+
+SEXP attribute_hidden do_curlGetHeaders(SEXP call, SEXP op, SEXP args, SEXP rho)
+{
+    if(!initialized) internet_Init();
+    if(initialized > 0)
+	return (*ptr->curlGetHeaders)(call, op, args, rho);
+    else {
+	error(_("internet routines cannot be loaded"));
+	return R_NilValue;
+    }
+}
+
+SEXP attribute_hidden do_curlDownload(SEXP call, SEXP op, SEXP args, SEXP rho)
+{
+    if(!initialized) internet_Init();
+    if(initialized > 0)
+	return (*ptr->curlDownload)(call, op, args, rho);
+    else {
+	error(_("internet routines cannot be loaded"));
+	return R_NilValue;
+    }
+}
+
+Rconnection attribute_hidden
+R_newCurlUrl(const char *description, const char * const mode, int type)
+{
+    if(!initialized) internet_Init();
+    if(initialized > 0)
+	return (*ptr->newcurlurl)(description, mode, type);
+    else {
+	error(_("internet routines cannot be loaded"));
+	return (Rconnection)0;
+    }
+    return (Rconnection)0; /* -Wall */
+}
+
