@@ -1,8 +1,8 @@
 /*
  *  Mathlib : A C Library of Special Functions
  *  Copyright (C) 1998 Ross Ihaka
- *  Copyright (C) 2000-2001 The R Core Team
- *  Copyright (C) 2005	The R Foundation
+ *  Copyright (C) 2000-2015 The R Core Team
+ *  Copyright (C) 2005-2015 The R Foundation
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -42,6 +42,7 @@
  */
 
 #include "nmath.h"
+#include "dpq.h"
 
 /* afc(i) :=  ln( i! )	[logarithm of the factorial i.
  *	   If (i > 7), use Stirling's approximation, otherwise use table lookup.
@@ -80,10 +81,8 @@ static double afc(int i)
 
 double rhyper(double nn1in, double nn2in, double kkin)
 {
-    const static double con = 57.56462733;
     const static double deltal = 0.0078;
     const static double deltau = 0.0034;
-    const static double scale = 1e25;
 
     /* extern double afc(int); */
 
@@ -91,29 +90,26 @@ double rhyper(double nn1in, double nn2in, double kkin)
     int i, ix;
     Rboolean reject, setup1, setup2;
 
-    double e, f, g, p, r, t, u, v, y;
-    double de, dg, dr, ds, dt, gl, gu, nk, nm, ub;
-    double xk, xm, xn, y1, ym, yn, yk, alv;
-
-    /* These should become `thread_local globals' : */
-    static int ks = -1;
-    static int n1s = -1, n2s = -1;
-
-    static int k, m;
-    static int minjx, maxjx, n1, n2;
-
-    static double a, d, s, w;
-    static double tn, xl, xr, kl, kr, lamdl, lamdr, p1, p2, p3;
+    double e, f, g, r, t, y;
 
+    /* These should become 'thread_local globals' : */
+    static int ks = -1, n1s = -1, n2s = -1;
+    static int k, m, minjx, maxjx, n1, n2;
+    static double tn;
+    // II :
+    static double w;
+    // III:
+    static double a, d, s, xl, xr, kl, kr, lamdl, lamdr, p1, p2, p3;
 
     /* check parameter validity */
 
     if(!R_FINITE(nn1in) || !R_FINITE(nn2in) || !R_FINITE(kkin))
 	ML_ERR_return_NAN;
 
-    nn1 = floor(nn1in+0.5);
-    nn2 = floor(nn2in+0.5);
-    kk	= floor(kkin +0.5);
+    // Disabling large (nn1, nn2, kk) { =^= rhyper (m,n,k) }:
+    nn1 = (int) R_forceint(nn1in);
+    nn2 = (int) R_forceint(nn2in);
+    kk	= (int) R_forceint(kkin);
 
     if (nn1 < 0 || nn2 < 0 || kk < 0 || kk > nn1 + nn2)
 	ML_ERR_return_NAN;
@@ -142,13 +138,13 @@ double rhyper(double nn1in, double nn2in, double kkin)
     if (setup2) {
 	ks = kk;
 	if (kk + kk >= tn) {
-	    k = tn - kk;
+	    k = (int)(tn - kk);
 	} else {
 	    k = kk;
 	}
     }
     if (setup1 || setup2) {
-	m = (k + 1.0) * (n1 + 1.0) / (tn + 2.0);
+	m = (int) ((k + 1.0) * (n1 + 1.0) / (tn + 2.0));
 	minjx = imax2(0, k - n2);
 	maxjx = imin2(n1, k);
     }
@@ -172,31 +168,35 @@ double rhyper(double nn1in, double nn2in, double kkin)
 	}
 	return ix;
 
-    } else if (m - minjx < 10) { /* II: inverse transformation ---------- */
+    } else if (m - minjx < 10) { // II: (Scaled) algorithm HIN (inverse transformation) ----
+	const static double scale = 1e25; // scaling factor against (early) underflow
+	const static double con = 57.5646273248511421;
+                                          // 25*log(10) = log(scale) { <==> exp(con) == scale }
 	if (setup1 || setup2) {
+	    double lw; // log(w);  w = exp(lw) * scale = exp(lw + log(scale)) = exp(lw + con)
 	    if (k < n2) {
-		w = exp(con + afc(n2) + afc(n1 + n2 - k)
-			- afc(n2 - k) - afc(n1 + n2));
+		lw = afc(n2) + afc(n1 + n2 - k) - afc(n2 - k) - afc(n1 + n2);
 	    } else {
-		w = exp(con + afc(n1) + afc(k)
-			- afc(k - n2) - afc(n1 + n2));
+		lw = afc(n1) + afc(     k     ) - afc(k - n2) - afc(n1 + n2);
 	    }
+	    w = exp(lw + con);
 	}
+	double p, u;
       L10:
 	p = w;
 	ix = minjx;
 	u = unif_rand() * scale;
-      L20:
-	if (u > p) {
+	while (u > p) {
 	    u -= p;
-	    p *= (n1 - ix) * (k - ix);
+	    p *= ((double) n1 - ix) * (k - ix);
 	    ix++;
 	    p = p / ix / (n2 - k + ix);
 	    if (ix > maxjx)
 		goto L10;
-	    goto L20;
 	}
-    } else { /* III : h2pe --------------------------------------------- */
+    } else { /* III : H2PE Algorithm --------------------------------------- */
+
+	double u,v;
 
 	if (setup1 || setup2) {
 	    s = sqrt((tn - k) * k * n1 * n2 / (tn - 1) / tn / tn);
@@ -225,14 +225,14 @@ double rhyper(double nn1in, double nn2in, double kkin)
 	u = unif_rand() * p3;
 	v = unif_rand();
 	if (u < p1) {		/* rectangular region */
-	    ix = xl + u;
+	    ix = (int) (xl + u);
 	} else if (u <= p2) {	/* left tail */
-	    ix = xl + log(v) / lamdl;
+	    ix = (int) (xl + log(v) / lamdl);
 	    if (ix < minjx)
 		goto L30;
 	    v = v * (u - p1) * lamdl;
 	} else {		/* right tail */
-	    ix = xr - log(v) / lamdr;
+	    ix = (int) (xr - log(v) / lamdr);
 	    if (ix > maxjx)
 		goto L30;
 	    v = v * (u - p2) * lamdr;
@@ -259,6 +259,8 @@ double rhyper(double nn1in, double nn2in, double kkin)
 		reject = FALSE;
 	    }
 	} else {
+	    double de, dg, dr, ds, dt, gl, gu, nk, nm, ub;
+	    double xk, xm, xn, y1, ym, yn, yk, alv;
 	    /* squeeze using upper and lower bounds */
 	    y = ix;
 	    y1 = y + 1.0;
@@ -318,7 +320,7 @@ double rhyper(double nn1in, double nn2in, double kkin)
 		    }
 		}
 	    }
-	}
+	} // else
 	if (reject)
 	    goto L30;
     }
